"""Wavefront implementation of a histogram.

@author: Hao Song (songhao@vmware.com)
"""
from __future__ import division
# pylint: disable=E0012,R0205

import math
import threading
import time

import tdigest


class Snapshot(object):
    """Wrapper for TDigest distribution."""

    def __init__(self, dist):
        """Construct TDigest Wrapper.

        @param dist: tdigest.TDigest
        @type dist: tdigest.TDigest
        """
        self.distribution = dist

    def get_max(self):
        """Get the maximum value in the distribution.

        @return: Maximum value
        """
        try:
            max_val = self.distribution.percentile(100)
        except ValueError:
            max_val = None
        return max_val

    def get_min(self):
        """Get the minimum value in the distribution.

        @return: minimum value
        """
        try:
            min_val = self.distribution.percentile(0)
        except ValueError:
            min_val = None
        return min_val

    def get_mean(self):
        """Get the mean of the values in the distribution.

        @return: mean value
        """
        try:
            return self.distribution.trimmed_mean(0, 100)
        except ZeroDivisionError:
            return None
        # Equivalent approach:
        # centroids = self.distribution.centroids_to_list()
        # if not centroids:
        #     return None
        # return sum(c['c'] * c['m'] for c in centroids) / sum(c['c'] for c in
        #                                                      centroids)

    def get_sum(self):
        """Get the sum of the values in the distribution.

        @return: sum of value
        """
        centroids = self.distribution.centroids_to_list()
        return sum(c['c'] * c['m'] for c in centroids)

    def get_value(self, quantile):
        """Get the value of given quantile.

        @param quantile: Quantile, range from 0 to 1
        @type quantile: float
        @return: the value in the distribution at the given quantile.
        Return None if distribution is empty.
        """
        percentile = quantile * 100
        try:
            return self.distribution.percentile(percentile)
        except ValueError:
            # ValueError("Tree is empty") from TDigest
            return None

    def get_size(self):
        """Get the size of snapshot.

        @return: Size of snapshot.
        """
        return self.distribution.n

    def get_count(self):
        """Get the number of values in the distribution.

        @return: Number of values
        """
        return self.get_size()


class Distribution(object):
    """Representation of a histogram distribution.

    Containing a timestamp and a list of centroids.
    """

    # pylint: disable=too-few-public-methods
    def __init__(self, timestamp, centroids):
        """Construct a distribution.

        @param timestamp: Timestamp in milliseconds since the epoch.
        @type timestamp: long
        @param centroids: Centroids
        @type centroids: list of tuple
        """
        self.timestamp = timestamp
        self.centroids = centroids


class ThreadMinuteBin(object):
    """Representation of a bin holds all threads histogram data in a min."""

    # pylint: disable=too-few-public-methods
    def __init__(self, accuracy=100, minute_millis=None):
        """Construct the Minute Bin.

        @param accuracy: Accuracy range from [0, 100]
        @type accuracy: int
        @param minute_millis: The timestamp at the start of the minute.
        @type minute_millis: long
        """
        self.accuracy = accuracy
        self.minute_millis = minute_millis
        self.per_thread_dist = {}

    def get_dist_by_thread_id(self, thread_id):
        """Retrieve the thread-local dist in one given minute."""
        if thread_id not in self.per_thread_dist:
            self.per_thread_dist[thread_id] = tdigest.TDigest(delta=1 /
                                                              self.accuracy)
        return self.per_thread_dist[thread_id]

    def update_dist_by_thread_id(self, thread_id, value):
        """Update the value in the distribution of given thread id."""
        self.get_dist_by_thread_id(thread_id).update(value)

    def bulk_update_dist_by_thread_id(self, thread_id, means, counts):
        """Bulk update values in the distribution of given thread id."""
        if means and counts:
            current_bin = self.get_dist_by_thread_id(thread_id)
            for i in range(min(len(means), len(counts))):
                current_bin.update(means[i], counts[i])

    def get_centroids(self):
        """Get list of centroids for dists of all threads in this minute."""
        centroids = []
        for thread_id in self.per_thread_dist:
            centroids.extend(
                self.per_thread_dist[thread_id].centroids_to_list())
        return centroids

    def to_distribution(self):
        """Convert to Distribution."""
        distributions = []
        for thread_id in self.per_thread_dist:
            centroids = [(centroid['m'], int(centroid['c'])) for centroid in
                         self.per_thread_dist[thread_id].centroids_to_list()]
            distributions.append(
                Distribution(self.minute_millis, centroids))
        return distributions


class WavefrontHistogramImpl(object):
    """Wavefront implementation of a histogram."""

    # Accuracy = Compression = 1 / Delta.
    _ACCURACY = 100

    # If a thread's bin queue has exceeded MAX_BINS number of bins (e.g.,
    # the thread has data that has yet to be reported for more than MAX_BINS
    # number of minutes), delete the oldest bin. Defaulted to 10 because we
    # can expect the histogram to be reported at least once every 10 minutes.
    _MAX_BINS = 10

    def __init__(self, clock_millis=None):
        """Construct Wavefront Histogram.

        @param clock_millis: A function which returns timestamp.
        @type clock_millis: function
        """
        self._clock_millis = clock_millis or self.current_clock_millis
        self._prior_minute_bins_list = []
        self._lock = threading.Lock()
        self._current_minute_bin = ThreadMinuteBin(
            self._ACCURACY, self.current_minute_millis())

    @staticmethod
    def current_clock_millis():
        """Get current time in millisecond."""
        return time.time() * 1000

    def current_minute_millis(self):
        """Get the timestamp of start of certain minute."""
        return int(self._clock_millis() / 60000) * 60000

    def update(self, value):
        """Add one value to the distribution.

        @param value: value to add.
        @type value: float
        """
        self.get_current_bin().update_dist_by_thread_id(
            threading.current_thread().ident, value)

    def bulk_update(self, means, counts):
        """Bulk-update this histogram with a set of centroids.

        @param means: the centroid values
        @type means: list of float
        @param counts: the centroid weights/sample counts
        @type counts: list of int
        """
        self.get_current_bin().bulk_update_dist_by_thread_id(
            threading.current_thread().ident, means, counts)

    def get_current_bin(self):
        """Retrieve the current bin.

        Will be invoked on the thread local histogram_bins_list.
        @return: Current minute bin
        @rtype: ThreadMinuteBin
        """
        return self.get_or_update_current_bin(self.current_minute_millis())

    def get_or_update_current_bin(self, curr_minute_millis):
        """Get current minute bin.

        Will update _prior_minute_bins_list if the minute has passed.
        @param curr_minute_millis: Current minute in millis
        @return: ThreadMinuteBin of current minute.
        @rtype: ThreadMinuteBin
        """
        if self._current_minute_bin.minute_millis == curr_minute_millis:
            return self._current_minute_bin
        # only one update thread can flush the current bin to
        # _prior_minute_bins_list and update the current bin.
        with self._lock:
            # Double check the minute millis of current bin to avoid if there
            # are multiple threads entering this block and the first thread
            # already did the flushing.
            if self._current_minute_bin.minute_millis != curr_minute_millis:
                if len(self._prior_minute_bins_list) > self._MAX_BINS:
                    del self._prior_minute_bins_list[0]
                self._prior_minute_bins_list.append(self._current_minute_bin)
                self._current_minute_bin = ThreadMinuteBin(self._ACCURACY,
                                                           curr_minute_millis)
            return self._current_minute_bin

    def get_prior_minute_bins_list(self):
        """Return newly-updated _prior_minute_bins_list."""
        self.get_or_update_current_bin(self.current_minute_millis())
        return self._prior_minute_bins_list

    def std_dev(self):
        """Return the stdDev of the values in the distribution."""
        mean = self.get_mean()
        variance_sum = 0
        count = 0
        for minute_bin in self.get_prior_minute_bins_list():
            centroids = minute_bin.get_centroids()
            count += sum(c['c'] for c in centroids)
            variance_sum += sum(c['c'] * ((c['m'] - mean) ** 2) for c in
                                centroids)
        variance = 0 if count == 0 else variance_sum / count
        return math.sqrt(variance)

    def flush_distributions(self):
        """Aggregate all the minute bins prior to the current minute.

        Only aggregating before current minute is because threads might be
        updating the current minute bin while the method is invoked.

        Returns a list of the distributions held within each bin.

        Note that invoking this method will also clear all data from the
        aggregated bins, thereby changing the state of the system and
        preventing data from being flushed more than once.

        @return: list of the distributions held within each bin
        @rtype: list of Distribution
        """
        distributions = []
        self.get_or_update_current_bin(self.current_minute_millis())
        with self._lock:
            for minute_bin in self._prior_minute_bins_list[:]:
                distributions.extend(minute_bin.to_distribution())
                self._prior_minute_bins_list.remove(minute_bin)
        return distributions

    def get_count(self):
        """Get the number of values in the distribution."""
        count = 0
        for minute_bin in self.get_prior_minute_bins_list():
            if minute_bin.per_thread_dist:
                count += sum(dist.n for dist
                             in minute_bin.per_thread_dist.values())
        return count

    def get_sum(self):
        """Get the sum of the values in the distribution."""
        res = 0
        for minute_bin in self.get_prior_minute_bins_list():
            res += sum(c['c'] * c['m'] for c in minute_bin.get_centroids())
        return res

    def get_snapshot(self):
        """Return a statistical of the histogram distribution.

        @return: Snapshot of Histogram
        @rtype: Snapshot
        """
        snapshot = tdigest.TDigest(1 / self._ACCURACY)
        for minute_bin in self.get_prior_minute_bins_list():
            for thread_id in minute_bin.per_thread_dist:
                snapshot = snapshot + minute_bin.per_thread_dist[thread_id]
        return Snapshot(snapshot)

    def get_max(self):
        """Get the maximum value in the distribution.

        Return None if the distribution is empty.
        """
        max_val = float('-inf')
        for minute_bin in self.get_prior_minute_bins_list():
            if minute_bin.per_thread_dist:
                max_val = max(max(dist.percentile(100) for dist
                                  in minute_bin.per_thread_dist.values()),
                              max_val)
        return None if float('-inf') == max_val else max_val

    def get_min(self):
        """Get the minimum value in the distribution.

        Return None if the distribution is empty.
        """
        min_val = float('inf')
        for minute_bin in self.get_prior_minute_bins_list():
            if minute_bin.per_thread_dist:
                min_val = min(min(dist.percentile(0) for dist
                                  in minute_bin.per_thread_dist.values()),
                              min_val)
        return None if float('inf') == min_val else min_val

    def get_mean(self):
        """Get the mean of the values in the distribution.

        Return None if the distribution is empty.
        """
        count = total = 0
        for minute_bin in self.get_prior_minute_bins_list():
            centroids = minute_bin.get_centroids()
            count += sum(c['c'] for c in centroids)
            total += sum(c['c'] * c['m'] for c in centroids)
        if count == 0:
            return None
        return total / count



# -*- coding: utf-8 -*-
"""Span Log.

@author Hao Song (songhao@vmware.com)`
"""


# pylint: disable=E0012,R0205
class SpanLog(object):
    """Span Log including timestamp and fields."""

    # pylint: disable=too-few-public-methods
    def __init__(self, timestamp, fields):
        """
        Construct Span Log.

        @param timestamp: Timestamp of the span log.
        @param fields: Fields of the span log.
        """
        self.timestamp = timestamp
        self.fields = fields



# -*- coding: utf-8 -*-
"""Wavefront Direct Ingestion Client.

@author Hao Song (songhao@vmware.com)
"""

from __future__ import absolute_import

import logging
import socket
import threading

import requests

try:
    import queue
except ImportError:
    import Queue as queue  # noqa

from . import entities
from .common import connection_handler, constants, utils
from .common.metrics import registry


# pylint: disable=too-many-instance-attributes
class WavefrontDirectClient(connection_handler.ConnectionHandler,
                            entities.WavefrontMetricSender,
                            entities.WavefrontHistogramSender,
                            entities.WavefrontTracingSpanSender,
                            entities.WavefrontEventSender):
    """Wavefront direct ingestion client.

    Sends data directly to Wavefront cluster via the direct ingestion API.
    """

    WAVEFRONT_METRIC_FORMAT = 'wavefront'
    WAVEFRONT_HISTOGRAM_FORMAT = 'histogram'
    WAVEFRONT_TRACING_SPAN_FORMAT = 'trace'
    WAVEFRONT_SPAN_LOG_FORMAT = 'spanLogs'
    WAVEFRONT_EVENT_FORMAT = 'event'

    REPORT_END_POINT = '/report'
    EVENT_END_POINT = '/api/v2/event'

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-statements

    def __init__(self,
                 server,
                 token,
                 max_queue_size=50000,
                 batch_size=10000,
                 flush_interval_seconds=5,
                 enable_internal_metrics=True):
        """Construct Direct Client.

        @param server: Server address, Example: https://INSTANCE.wavefront.com
        @type server: str
        @param token: Token with Direct Data Ingestion permission granted
        @type token: str
        @param max_queue_size:
        @type max_queue_size: int
        Max Queue Size, size of internal data buffer for each data type.
        50000 by default
        @param batch_size:
        @type batch_size: int
        Batch Size, amount of data sent by one api call, 10000 by default
        @param flush_interval_seconds: Interval flush time, 5 secs by default
        @type flush_interval_seconds: int
        """
        super().__init__()
        self.server = server
        self._token = token
        self._max_queue_size = max_queue_size
        self._batch_size = batch_size
        self._flush_interval_seconds = flush_interval_seconds
        self._default_source = socket.gethostname() or 'unknown'
        self._metrics_buffer = queue.Queue(max_queue_size)
        self._histograms_buffer = queue.Queue(max_queue_size)
        self._tracing_spans_buffer = queue.Queue(max_queue_size)
        self._spans_log_buffer = queue.Queue(max_queue_size)
        self._events_buffer = queue.Queue(max_queue_size)
        self._headers = {'Content-Type': 'application/octet-stream',
                         'Content-Encoding': 'gzip',
                         'Authorization': 'Bearer ' + token}
        self._event_headers = {'Content-Type': 'application/json',
                               'Content-Encoding': 'gzip',
                               'Authorization': 'Bearer ' + token}
        self._closed = False
        self._schedule_lock = threading.Lock()
        self._timer = None
        self._schedule_timer()

        if enable_internal_metrics:
            self._sdk_metrics_registry = registry.WavefrontSdkMetricsRegistry(
                wf_metric_sender=self,
                prefix='{}.core.sender.direct'.format(
                    constants.SDK_METRIC_PREFIX))
        else:
            self._sdk_metrics_registry = registry.WavefrontSdkMetricsRegistry(
                wf_metric_sender=None)

        semver = utils.get_sem_ver(constants.WAVEFRONT_SDK_PYTHON)

        def version():
            return semver
        self._sdk_metrics_registry.new_gauge('version', version)
        self._sdk_metrics_registry.new_gauge(
            'points.queue.size', self._metrics_buffer.qsize)
        self._sdk_metrics_registry.new_gauge(
            'points.queue.remaining_capacity',
            remaining_capacity_getter(self._metrics_buffer))
        self._points_valid = self._sdk_metrics_registry.new_counter(
            'points.valid')
        self._points_invalid = self._sdk_metrics_registry.new_counter(
            'points.invalid')
        self._points_dropped = self._sdk_metrics_registry.new_counter(
            'points.dropped')
        self._points_report_errors = self._sdk_metrics_registry.new_counter(
            'points.report.errors')

        self._sdk_metrics_registry.new_gauge(
            'histograms.queue.size', self._histograms_buffer.qsize)
        self._sdk_metrics_registry.new_gauge(
            'histograms.queue.remaining_capacity',
            remaining_capacity_getter(self._histograms_buffer))
        self._histograms_valid = self._sdk_metrics_registry.new_counter(
            'histograms.valid')
        self._histograms_invalid = self._sdk_metrics_registry.new_counter(
            'histograms.invalid')
        self._histograms_dropped = self._sdk_metrics_registry.new_counter(
            'histograms.dropped')
        self._histograms_report_errors = (
            self._sdk_metrics_registry.new_counter('histograms.report.errors'))

        self._sdk_metrics_registry.new_gauge(
            'spans.queue.size', self._tracing_spans_buffer.qsize)
        self._sdk_metrics_registry.new_gauge(
            'spans.queue.remaining_capacity',
            remaining_capacity_getter(self._tracing_spans_buffer))
        self._spans_valid = self._sdk_metrics_registry.new_counter(
            'spans.valid')
        self._spans_invalid = self._sdk_metrics_registry.new_counter(
            'spans.invalid')
        self._spans_dropped = self._sdk_metrics_registry.new_counter(
            'spans.dropped')
        self._spans_report_errors = self._sdk_metrics_registry.new_counter(
            'spans.report.errors')

        self._sdk_metrics_registry.new_gauge(
            'span_logs.queue.size', self._spans_log_buffer.qsize)
        self._sdk_metrics_registry.new_gauge(
            'span_logs.queue.remaining_capacity',
            remaining_capacity_getter(self._spans_log_buffer))
        self._span_logs_valid = self._sdk_metrics_registry.new_counter(
            'span_logs.valid')
        self._span_logs_invalid = self._sdk_metrics_registry.new_counter(
            'span_logs.invalid')
        self._span_logs_dropped = self._sdk_metrics_registry.new_counter(
            'span_logs.dropped')
        self._span_logs_report_errors = self._sdk_metrics_registry.new_counter(
            'span_logs.report.errors')

        self._sdk_metrics_registry.new_gauge(
            'events.queue.size', self._events_buffer.qsize)
        self._sdk_metrics_registry.new_gauge(
            'events.queue.remaining_capacity',
            remaining_capacity_getter(self._events_buffer))
        self._events_valid = self._sdk_metrics_registry.new_counter(
            'events.valid')
        self._events_invalid = self._sdk_metrics_registry.new_counter(
            'events.invalid')
        self._events_dropped = self._sdk_metrics_registry.new_counter(
            'events.dropped')
        self._events_report_errors = self._sdk_metrics_registry.new_counter(
            'events.report.errors')

    def _report(self, points, data_format, entity_prefix, report_errors):
        r"""One api call sending one given string data.

        @param points: List of data in string format, concat by '\n'
        @type points: str
        @param data_format: Type of data to be sent
        @type data_format: str
        """
        try:
            if data_format == self.WAVEFRONT_EVENT_FORMAT:
                response = requests.post(self.server + self.EVENT_END_POINT,
                                         params=None,
                                         headers=self._event_headers,
                                         data=points)
            else:
                params = {'f': data_format}
                compressed_data = utils.gzip_compress(points.encode('utf-8'))
                response = requests.post(self.server + self.REPORT_END_POINT,
                                         params=params,
                                         headers=self._headers,
                                         data=compressed_data)

            self._sdk_metrics_registry.new_counter(
                '{}.report.{}'.format(entity_prefix,
                                      response.status_code)).inc()
            response.raise_for_status()
        except Exception as error:
            report_errors.inc()
            raise error

    def _batch_report(self, batch_line_data, data_format, entity_prefix,
                      report_errors):
        """One api call sending one given list of data.

        @param batch_line_data: List of data to be sent
        @type batch_line_data: list
        @param data_format: Type of data to be sent
        @type data_format: str
        """
        # Sending events through direct ingestion does not support batching.
        batch_size = int(
            data_format == self.WAVEFRONT_EVENT_FORMAT) or self._batch_size

        # Split data into chunks, each with the size of given batch_size
        for batch in utils.chunks(batch_line_data, batch_size):
            # report once per batch
            try:
                self._report('\n'.join(batch) + '\n', data_format,
                             entity_prefix, report_errors)
            # pylint: disable=broad-except,fixme
            # TODO: Please replace a generic Exception with a specific one.
            except Exception as error:
                logging.error(
                    'Failed to report %s data points to wavefront %s',
                    data_format, error)

    def _internal_flush(self, data_buffer, data_format, entity_prefix,
                        report_errors):
        """Get all data from one data buffer to a list, and report that list.

        @param data_buffer: Data buffer to be flush and sent
        @type: Queue
        @param data_format: Type of data to be sent
        @type: str
        """
        data = []
        size = data_buffer.qsize()
        while size > 0 and not data_buffer.empty():
            data.append(data_buffer.get())
            size -= 1
        self._batch_report(data, data_format, entity_prefix, report_errors)

    def _schedule_timer(self):
        # Flush every 5 secs by default
        if not self._closed:
            self._timer = threading.Timer(self._flush_interval_seconds,
                                          self._flush)
            self._timer.daemon = True
            self._timer.start()

    def _flush(self):
        """Use Timer to keep flushing each <flush_interval_seconds> secs."""
        try:
            self.flush_now()
        finally:
            with self._schedule_lock:
                if not self._closed:
                    self._schedule_timer()

    def flush_now(self):
        """Flush all the data buffer immediately."""
        self._internal_flush(self._metrics_buffer,
                             self.WAVEFRONT_METRIC_FORMAT, 'points',
                             self._points_report_errors)
        self._internal_flush(self._histograms_buffer,
                             self.WAVEFRONT_HISTOGRAM_FORMAT, 'histograms',
                             self._histograms_report_errors)
        self._internal_flush(self._tracing_spans_buffer,
                             self.WAVEFRONT_TRACING_SPAN_FORMAT, 'spans',
                             self._spans_report_errors)
        self._internal_flush(self._spans_log_buffer,
                             self.WAVEFRONT_SPAN_LOG_FORMAT, 'span_logs',
                             self._span_logs_report_errors)
        self._internal_flush(self._events_buffer,
                             self.WAVEFRONT_EVENT_FORMAT, 'events',
                             self._events_report_errors)

    def close(self):
        """Flush all buffer before close the client."""
        self.flush_now()
        with self._schedule_lock:
            self._closed = True
            if self._timer is not None:
                self._timer.cancel()
        self._sdk_metrics_registry.close(timeout_secs=1)

    def send_metric(self, name, value, timestamp, source, tags):
        """Send Metric Data via direct ingestion client.

        Wavefront Metrics Data format
        <metricName> <metricValue> [<timestamp>] source=<source> [pointTags]
        Example: "new-york.power.usage 42422 1533531013 source=localhost
                  datacenter=dc1"

        @param name: Metric Name
        @type name: str
        @param value: Metric Value
        @type value: float
        @param timestamp: Timestamp
        @type timestamp: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: dict
        """
        try:
            line_data = utils.metric_to_line_data(
                name, value, timestamp, source, tags, self._default_source)
            self._points_valid.inc()
        except ValueError as error:
            self._points_invalid.inc()
            raise error
        try:
            self._metrics_buffer.put(line_data)
        except queue.Full as error:
            self._points_dropped.inc()
            raise error

    def send_metric_now(self, metrics):
        """Send a list of metrics immediately.

        Have to construct the data manually by calling
        common.utils.metric_to_line_data()

        @param metrics: List of string spans data
        @type metrics: list[str]
        """
        self._batch_report(metrics, self.WAVEFRONT_METRIC_FORMAT, 'points',
                           self._points_report_errors)

    def send_distribution(self, name, centroids, histogram_granularities,
                          timestamp, source, tags):
        """Send Distribution Data via direct ingestion client.

        Wavefront Histogram Data format
        {!M | !H | !D} [<timestamp>] #<count> <mean> [centroids]
        <histogramName> source=<source> [pointTags]
        Example: "!M 1533531013 #20 30.0 #10 5.1 request.latency
                  source=appServer1 region=us-west"

        @param name: Histogram Name
        @type name: str
        @param centroids: List of centroids(pairs)
        @type centroids: list
        @param histogram_granularities: Histogram Granularities
        @type histogram_granularities: set
        @param timestamp: Timestamp
        @type timestamp: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: dict
        """
        try:
            line_data = utils.histogram_to_line_data(
                name, centroids, histogram_granularities, timestamp,
                source, tags, self._default_source)
            self._histograms_valid.inc()
        except ValueError as error:
            self._histograms_invalid.inc()
            raise error
        try:
            self._histograms_buffer.put(line_data)
        except queue.Full as error:
            self._histograms_dropped.inc()
            raise error

    def send_distribution_now(self, distributions):
        """Send a list of distribution immediately.

        Have to construct the data manually by calling
        common.utils.histogram_to_line_data()

        @param distributions: List of string spans data
        @type distributions: list[str]
        """
        self._batch_report(distributions, self.WAVEFRONT_HISTOGRAM_FORMAT,
                           'histograms', self._histograms_report_errors)

    # pylint: disable=too-many-arguments

    def send_span(self, name, start_millis, duration_millis, source, trace_id,
                  span_id, parents, follows_from, tags, span_logs):
        """Send span data via direct ingestion client.

        Wavefront Tracing Span Data format
        <tracingSpanName> source=<source> [pointTags] <start_millis>
        <duration_milli_seconds>
        Example: "getAllUsers source=localhost
                  traceId=7b3bf470-9456-11e8-9eb6-529269fb1459
                  spanId=0313bafe-9457-11e8-9eb6-529269fb1459
                  parent=2f64e538-9457-11e8-9eb6-529269fb1459
                  application=Wavefront http.method=GET
                  1533531013 343500"

        @param name: Span Name
        @type name: str
        @param start_millis: Start time
        @type start_millis: long
        @param duration_millis: Duration time
        @type duration_millis: long
        @param source: Source
        @type source: str
        @param trace_id: Trace ID
        @type trace_id: UUID
        @param span_id: Span ID
        @type span_id: UUID
        @param parents: Parents Span ID
        @type parents: List of UUID
        @param follows_from: Follows Span ID
        @type follows_from: List of UUID
        @param tags: Tags
        @type tags: list
        @param span_logs: Span Log
        """
        try:
            line_data = utils.tracing_span_to_line_data(
                name, start_millis, duration_millis, source, trace_id, span_id,
                parents, follows_from, tags, span_logs, self._default_source)
            self._spans_valid.inc()
        except ValueError as error:
            self._spans_invalid.inc()
            raise error
        try:
            self._tracing_spans_buffer.put(line_data)
        except queue.Full as error:
            self._spans_dropped.inc()
            raise error
        if span_logs:
            try:
                line_data = utils.span_log_to_line_data(
                    trace_id, span_id, span_logs)
                self._span_logs_valid.inc()
            except ValueError as error:
                self._span_logs_invalid.inc()
                raise error
            try:
                self._spans_log_buffer.put(line_data)
            except queue.Full as error:
                self._span_logs_dropped.inc()
                raise error

    def send_span_now(self, spans):
        """
        Send a list of spans immediately.

        Have to construct the data manually by calling
        common.utils.tracing_span_to_line_data()

        @param spans: List of string spans data
        @type spans: list[str]
        """
        self._batch_report(spans, self.WAVEFRONT_TRACING_SPAN_FORMAT, 'spans',
                           self._spans_report_errors)

    def send_span_log_now(self, span_logs):
        """
        Send a list of spans logs immediately.

        Have to construct the data manually by calling
        common.utils.span_log_to_line_data()

        @param span_logs: List of string span logs data
        @type span_logs: list[str]
        """
        self._batch_report(span_logs, self.WAVEFRONT_SPAN_LOG_FORMAT,
                           'span_logs', self._span_logs_report_errors)

    def send_event(self, name, start_time, end_time, source, tags,
                   annotations):
        """Send Event Data via direct ingestion client.

        Wavefront Event Data format
        {"name": <Event Name>, "annotations": <Annotations>,
         "hosts": <Host Name>,"startTime": <Start Time>,
          "endTime": <End Time>, "tags": <Tags>}
        Example: {"name": event_via_direct_ingestion, "annotations": {
        "severity": "severe", "type": "backup", "details": "broker backup"},
         "hosts": ["localhost"], "startTime": 1590678089,
         "endTime": 1590679089, "tags": ["env:", "test"]}

        @param name: Event Name
        @type name: str
        @param start_time: Event Start Time
        @type start_time: long
        @param end_time: Event End Time
        @type end_time: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: list[str]
        @param annotations: Annotations
        @type annotations: dict
        """
        try:
            line_data = utils.event_to_json(
                name, start_time, end_time, source, tags, annotations,
                self._default_source)
            self._events_valid.inc()
        except ValueError as error:
            self._events_invalid.inc()
            raise error
        try:
            self._events_buffer.put(line_data)
        except queue.Full as error:
            self._events_dropped.inc()
            raise error

    def send_event_now(self, events):
        """Send a list of events immediately.

        Have to construct the data manually by calling
        common.utils.event_to_json()

        @param events: List of string events data
        @type events: list[str]
        """
        self._batch_report(events, self.WAVEFRONT_EVENT_FORMAT, 'events',
                           self._events_report_errors)

    def get_failure_count(self):
        """Get failure count for one connection."""
        return (self._points_report_errors.count() +
                self._histograms_report_errors.count() +
                self._spans_report_errors.count() +
                self._span_logs_report_errors.count() +
                self._events_report_errors.count())


def remaining_capacity_getter(buf):
    """Get remaining capacity of given queue.

    @param buf: Input Buffer Queue.
    @type buf: queue.Queue
    @return: Remaining capacity of the queue.
    @rtype: int
    """
    return lambda: buf.maxsize - buf.qsize()



# -*- coding: utf-8 -*-
"""Wavefront Proxy Client.

@author Hao Song (songhao@vmware.com).
"""

from __future__ import absolute_import

import logging
from socket import gethostname

from . import entities
from .common import constants, utils
from .common.metrics import registry
from .common.proxy_connection_handler import ProxyConnectionHandler

LOGGER = logging.getLogger('wavefront_sdk.WavefrontDirectClient')


# pylint: disable=too-many-instance-attributes
class WavefrontProxyClient(entities.WavefrontMetricSender,
                           entities.WavefrontHistogramSender,
                           entities.WavefrontTracingSpanSender):
    """WavefrontProxyClient that sends data directly via TCP.

    User should probably attempt to reconnect
    when exceptions are thrown from any methods.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, host, metrics_port, distribution_port, tracing_port,
                 event_port=None, timeout=None, enable_internal_metrics=True):
        """Construct Proxy Client.

        @param host: Hostname of the Wavefront proxy, 2878 by default
        @param metrics_port:
        Metrics Port on which the Wavefront proxy is listening on
        @param distribution_port:
        Distribution Port on which the Wavefront proxy is listening on
        @param tracing_port:
        Tracing Port on which the Wavefront proxy is listening on
        @param event_port:
        Event Port on which the Wavefront proxy is listening on
        @param timeout:
        Timeout to set on socket connecting to proxy
        @param enable_internal_metrics
        Flag to enable/disable internal metrics
        """
        self.proxy_host = host
        self.metrics_port = metrics_port
        self.distribution_port = distribution_port
        self.tracing_port = tracing_port
        self.event_port = event_port

        if enable_internal_metrics:
            self._sdk_metrics_registry = registry.WavefrontSdkMetricsRegistry(
                wf_metric_sender=self,
                prefix='{}.core.sender.proxy'.format(
                    constants.SDK_METRIC_PREFIX))
        else:
            self._sdk_metrics_registry = registry.WavefrontSdkMetricsRegistry(
                wf_metric_sender=None)

        self._metrics_proxy_connection_handler = (
            None if metrics_port is None
            else ProxyConnectionHandler(host, metrics_port,
                                        self._sdk_metrics_registry,
                                        'metricHandler',
                                        timeout=timeout))
        self._histogram_proxy_connection_handler = (
            None if distribution_port is None
            else ProxyConnectionHandler(host, distribution_port,
                                        self._sdk_metrics_registry,
                                        'histogramHandler',
                                        timeout=timeout))
        self._tracing_proxy_connection_handler = (
            None if tracing_port is None
            else ProxyConnectionHandler(host, tracing_port,
                                        self._sdk_metrics_registry,
                                        'tracingHandler',
                                        timeout=timeout))
        self._event_proxy_connection_handler = (
            None if event_port is None
            else ProxyConnectionHandler(host, event_port,
                                        self._sdk_metrics_registry,
                                        'eventHandler',
                                        timeout=timeout))
        self._default_source = gethostname()

        semver = utils.get_sem_ver(constants.WAVEFRONT_SDK_PYTHON)

        def version():
            return semver
        self._sdk_metrics_registry.new_gauge('version', version)
        self._points_discarded = self._sdk_metrics_registry.new_counter(
            'points.discarded')
        self._points_dropped = self._sdk_metrics_registry.new_counter(
            'points.dropped')
        self._points_valid = self._sdk_metrics_registry.new_counter(
            'points.valid')
        self._points_invalid = self._sdk_metrics_registry.new_counter(
            'points.invalid')

        self._histograms_dropped = self._sdk_metrics_registry.new_counter(
            'histograms.dropped')
        self._histograms_discarded = self._sdk_metrics_registry.new_counter(
            'histograms.discarded')
        self._histograms_valid = self._sdk_metrics_registry.new_counter(
            'histograms.valid')
        self._histograms_invalid = self._sdk_metrics_registry.new_counter(
            'histograms.invalid')

        self._spans_dropped = self._sdk_metrics_registry.new_counter(
            'spans.dropped')
        self._spans_discarded = self._sdk_metrics_registry.new_counter(
            'spans.discarded')
        self._spans_valid = self._sdk_metrics_registry.new_counter(
            'spans.valid')
        self._spans_invalid = self._sdk_metrics_registry.new_counter(
            'spans.invalid')

        self._span_logs_dropped = self._sdk_metrics_registry.new_counter(
            'span_logs.dropped')
        self._span_logs_discarded = self._sdk_metrics_registry.new_counter(
            'span_logs.discarded')
        self._span_logs_valid = self._sdk_metrics_registry.new_counter(
            'span_logs.valid')
        self._span_logs_invalid = self._sdk_metrics_registry.new_counter(
            'span_logs.invalid')

        self._events_discarded = self._sdk_metrics_registry.new_counter(
            'events.discarded')
        self._events_dropped = self._sdk_metrics_registry.new_counter(
            'events.dropped')
        self._events_valid = self._sdk_metrics_registry.new_counter(
            'events.valid')
        self._events_invalid = self._sdk_metrics_registry.new_counter(
            'events.invalid')

    def close(self):
        """Close all proxy connections."""
        if self._metrics_proxy_connection_handler:
            self._metrics_proxy_connection_handler.close()
        if self._histogram_proxy_connection_handler:
            self._histogram_proxy_connection_handler.close()
        if self._tracing_proxy_connection_handler:
            self._tracing_proxy_connection_handler.close()
        if self._event_proxy_connection_handler:
            self._event_proxy_connection_handler.close()
        self._sdk_metrics_registry.close(timeout_secs=1)

    def get_failure_count(self):
        """Get Total Failure Count for all connections.

        @return: Failure Count
        """
        failure_count = 0
        if self._metrics_proxy_connection_handler:
            failure_count += (
                self._metrics_proxy_connection_handler.get_failure_count())
        if self._histogram_proxy_connection_handler:
            failure_count += (
                self._histogram_proxy_connection_handler.get_failure_count())
        if self._tracing_proxy_connection_handler:
            failure_count += (
                self._tracing_proxy_connection_handler.get_failure_count())
        if self._event_proxy_connection_handler:
            failure_count += (
                self._event_proxy_connection_handler.get_failure_count())
        return failure_count

    def send_metric(self, name, value, timestamp, source, tags):
        """Send Metric Data via proxy.

        Wavefront Metrics Data format
        <metricName> <metricValue> [<timestamp>] source=<source> [pointTags]
        Example: "new-york.power.usage 42422 1533531013 source=localhost
                  datacenter=dc1"

        @param name: Metric Name
        @type name: str
        @param value: Metric Value
        @type value: float
        @param timestamp: Timestamp
        @type timestamp: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: dict
        """
        if not self._metrics_proxy_connection_handler:
            self._points_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "metrics port for Wavefront proxy.")
        try:
            line_data = utils.metric_to_line_data(
                name, value, timestamp, source, tags, self._default_source)
            self._points_valid.inc()
            self._metrics_proxy_connection_handler.send_data(line_data)
        except ValueError:
            self._points_invalid.inc()
        except Exception as error:
            self._points_dropped.inc()
            self._metrics_proxy_connection_handler.increment_failure_count()
            raise error

    def send_metric_now(self, metrics):
        """Send a list of metrics immediately.

        Have to construct the data manually by calling
        common.utils.metric_to_line_data()

        @param metrics: List of string spans data
        @type metrics: list[str]
        """
        if not self._metrics_proxy_connection_handler:
            self._points_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "metrics port for Wavefront proxy.")
        for metric in metrics:
            try:
                self._metrics_proxy_connection_handler.send_data(metric)
            except Exception as error:
                self._points_dropped.inc()
                self._metrics_proxy_connection_handler.increment_failure_count(
                )
                raise error

    def send_distribution(self, name, centroids, histogram_granularities,
                          timestamp, source, tags):
        """Send Distribution Data via proxy.

        Wavefront Histogram Data format
        {!M | !H | !D} [<timestamp>] #<count> <mean> [centroids]
        <histogramName> source=<source> [pointTags]
        Example: "!M 1533531013 #20 30.0 #10 5.1 request.latency
                  source=appServer1 region=us-west"

        @param name: Histogram Name
        @type name: str
        @param centroids: List of centroids(pairs)
        @type centroids: list
        @param histogram_granularities: Histogram Granularities
        @type histogram_granularities: set
        @param timestamp: Timestamp
        @type timestamp: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: dict
        """
        if not self._histogram_proxy_connection_handler:
            self._histograms_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "histogram distribution port for Wavefront proxy.")
        try:
            line_data = utils.histogram_to_line_data(
                name, centroids, histogram_granularities, timestamp, source,
                tags, self._default_source)
            self._histograms_valid.inc()
            self._histogram_proxy_connection_handler.send_data(line_data)
        except ValueError:
            self._histograms_invalid.inc()
        except Exception as error:
            self._histograms_dropped.inc()
            self._histogram_proxy_connection_handler.increment_failure_count()
            raise error

    def send_distribution_now(self, distributions):
        """Send a list of distributions immediately.

        Have to construct the data manually by calling
        common.utils.histogram_to_line_data()

        @param distributions: List of string distribution data
        @type distributions: list[str]
        """
        if not self._histogram_proxy_connection_handler:
            self._histograms_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "histogram distribution port for Wavefront proxy.")
        for distribution in distributions:
            try:
                self._histogram_proxy_connection_handler.send_data(
                    distribution)
            except Exception as error:
                self._histograms_dropped.inc()
                (self._histogram_proxy_connection_handler
                 .increment_failure_count())
                raise error

    # pylint: disable=too-many-arguments
    def send_span(self, name, start_millis, duration_millis, source, trace_id,
                  span_id, parents, follows_from, tags, span_logs):
        """Send span data via proxy.

        Wavefront Tracing Span Data format
        <tracingSpanName> source=<source> [pointTags] <start_millis>
        <duration_milli_seconds>
        Example: "getAllUsers source=localhost
                  traceId=7b3bf470-9456-11e8-9eb6-529269fb1459
                  spanId=0313bafe-9457-11e8-9eb6-529269fb1459
                  parent=2f64e538-9457-11e8-9eb6-529269fb1459
                  application=Wavefront http.method=GET
                  1533531013 343500"

        @param name: Span Name
        @type name: str
        @param start_millis: Start time
        @type start_millis: long
        @param duration_millis: Duration time
        @type duration_millis: long
        @param source: Source
        @type source: str
        @param trace_id: Trace ID
        @type trace_id: UUID
        @param span_id: Span ID
        @type span_id: UUID
        @param parents: Parents Span ID
        @type parents: List of UUID
        @param follows_from: Follows Span ID
        @type follows_from: List of UUID
        @param tags: Tags
        @type tags: list
        @param span_logs: Span Log
        """
        if not self._tracing_proxy_connection_handler:
            self._spans_discarded.inc()
            if span_logs:
                self._span_logs_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "tracing port for Wavefront proxy.")
        try:
            line_data = utils.tracing_span_to_line_data(
                name, start_millis, duration_millis, source, trace_id,
                span_id, parents, follows_from, tags, span_logs,
                self._default_source)
            self._spans_valid.inc()
            self._tracing_proxy_connection_handler.send_data(line_data)
        except ValueError:
            self._spans_invalid.inc()
        except Exception as error:
            self._spans_dropped.inc()
            self._tracing_proxy_connection_handler.increment_failure_count()
            raise error
        if span_logs:
            try:
                span_log_line_data = utils.span_log_to_line_data(
                    trace_id, span_id, span_logs)
                self._span_logs_valid.inc()
                self._tracing_proxy_connection_handler.send_data(
                    span_log_line_data)
            except ValueError:
                self._span_logs_invalid.inc()
            except Exception as error:
                self._span_logs_dropped.inc()
                (self._tracing_proxy_connection_handler.
                 increment_failure_count())
                raise error

    def send_span_now(self, spans):
        """Send a list of spans immediately.

        Have to construct the data manually by calling
        common.utils.tracing_span_to_line_data()

        @param spans: List of string tracing span data
        @type spans: list[str]
        """
        if not self._tracing_proxy_connection_handler:
            self._spans_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "tracing port for Wavefront proxy.")
        for span in spans:
            try:
                self._tracing_proxy_connection_handler.send_data(span)
            except Exception as error:
                self._spans_dropped.inc()
                self._tracing_proxy_connection_handler.increment_failure_count(
                    )
                raise error

    def send_span_log_now(self, span_logs):
        """Send a list of span logs immediately.

        Have to construct the data manually by calling
        common.utils.tracing_span_to_line_data()

        @param span_logs: List of string span log data
        @type span_logs: list[str]
        """
        if not self._tracing_proxy_connection_handler:
            self._span_logs_discarded.inc()
            LOGGER.warning("Can't send data to Wavefront. Please configure "
                           "tracing port for Wavefront proxy.")
        for span_log in span_logs:
            try:
                self._tracing_proxy_connection_handler.send_data(span_log)
            except Exception as error:
                self._span_logs_dropped.inc()
                self._tracing_proxy_connection_handler.increment_failure_count(
                )
                raise error

    def send_event(self, name, start_time, end_time, source, tags,
                   annotations):
        """Send Event Data via proxy.

        Wavefront Event Data format
        @Event start_time end_time name [annotations]  host=<source> [Tags]
        Example: @Event 1590650592 1590650692 "event_via_proxy" severity="info"
         type="backup" details="broker backup" host="localhost" tag="env: dev"

        @param name: Event Name
        @type name: str
        @param start_time: Event Start Time
        @type start_time: long
        @param end_time: Event End Time
        @type end_time: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: list[str]
        @param annotations: Annotations
        @type annotations: dict
        """
        if not self._event_proxy_connection_handler:
            self._events_discarded.inc()
            LOGGER.warning("Can't send event to Wavefront. Please configure "
                           "events port for Wavefront proxy.")
        else:
            try:
                line_data = utils.event_to_line_data(
                    name, start_time, end_time, source, tags, annotations,
                    self._default_source)
                self._events_valid.inc()
                self._event_proxy_connection_handler.send_data(line_data)
            except ValueError:
                self._events_invalid.inc()
            except Exception as error:
                self._events_dropped.inc()
                self._event_proxy_connection_handler.increment_failure_count()
                raise error

    def send_event_now(self, events):
        """Send a list of events immediately.

        Have to construct the data manually by calling
        common.utils.event_to_line_data()

        @param events: List of string events data
        @type events: list[str]
        """
        if not self._event_proxy_connection_handler:
            self._events_discarded.inc()
            LOGGER.warning("Can't send event to Wavefront. Please configure "
                           "events port for Wavefront proxy.")
        else:
            for event in events:
                try:
                    self._event_proxy_connection_handler.send_data(event)
                except Exception as error:
                    self._events_dropped.inc()
                    self._event_proxy_connection_handler\
                        .increment_failure_count()
                    raise error



"""Unit Tests for Wavefront Python SDK.

@author Hao Song (songhao@vmware.com)
"""

import unittest
import uuid

from wavefront_sdk.common.utils import get_sem_ver_value
from wavefront_sdk.common.utils import histogram_to_line_data
from wavefront_sdk.common.utils import metric_to_line_data
from wavefront_sdk.common.utils import sanitize, sanitize_value, \
                                        sanitize_without_quotes
from wavefront_sdk.common.utils import tracing_span_to_line_data
from wavefront_sdk.entities import histogram_granularity


class TestUtils(unittest.TestCase):
    """Test Functions of wavefront_sdk.common.utils."""

    def test_get_sem_ver_value(self):
        """Test wavefront_sdk.common.utils.get_sem_ver_value()."""
        self.assertEqual("0.0", get_sem_ver_value(""))
        self.assertEqual("1.0100", get_sem_ver_value("1.1.0"))
        self.assertEqual("1.0100", get_sem_ver_value("1.1.0-SNAPSHOT"))
        self.assertEqual("1.0101", get_sem_ver_value("1.1.1"))
        self.assertEqual("1.1001", get_sem_ver_value("1.10.1"))
        self.assertEqual("1.0110", get_sem_ver_value("1.1.10"))
        self.assertEqual("1.0001", get_sem_ver_value("1.0.1"))
        self.assertEqual("1.0010", get_sem_ver_value("1.0.10"))
        self.assertEqual("1.1010", get_sem_ver_value("1.10.10"))

    def test_sanitize(self):
        """Test wavefront_sdk.common.utils.sanitize()."""
        self.assertEqual('"hello"', sanitize('hello'))
        self.assertEqual('"hello-world"', sanitize('hello world'))
        self.assertEqual('"hello.world"', sanitize('hello.world'))
        self.assertEqual('"hello-world-"', sanitize('hello"world"'))
        self.assertEqual('"hello-world"', sanitize("hello'world"))
        self.assertEqual('"hello-world"', sanitize('hello/world'))
        self.assertEqual('"~component.heartbeat"',
                         sanitize('~component.heartbeat'))
        self.assertEqual('"-component.heartbeat"',
                         sanitize('!component.heartbeat'))
        self.assertEqual('"Δcomponent.heartbeat"',
                         sanitize('Δcomponent.heartbeat'))
        self.assertEqual('"∆component.heartbeat"',
                         sanitize('∆component.heartbeat'))

    def test_sanitize_without_quotes(self):
        """Test wavefront_sdk.common.utils.sanitize_without_quotes()."""
        self.assertEqual('hello-world', sanitize_without_quotes('hello world'))

    def test_sanitize_value(self):
        """Test wavefront_sdk.common.utils.sanitize_value()."""
        self.assertEqual('"hello"', sanitize_value("hello"))
        self.assertEqual('"hello world"', sanitize_value("hello world"))
        self.assertEqual('"hello.world"', sanitize_value("hello.world"))
        self.assertEqual('"hello\\"world\\""',
                         sanitize_value("hello\"world\""))
        self.assertEqual("\"hello'world\"", sanitize_value("hello'world"))
        self.assertEqual('"hello\\nworld"', sanitize_value("hello\nworld"))
        self.assertEqual('"localhost:8080"', sanitize_value('localhost:8080'))
        self.assertEqual('"\\"127.0.0.1:8080\\""',
                         sanitize_value('"127.0.0.1:8080"'))

    def test_metric_to_line_data(self):
        """Test wavefront_sdk.common.utils.metric_to_line_data()."""
        self.assertEqual(
            '"new-york.power.usage" 42422.0 1493773500 source="localhost"'
            ' "datacenter"="dc1"\n',
            metric_to_line_data('new-york.power.usage', 42422, 1493773500,
                                'localhost', {'datacenter': 'dc1'},
                                'defaultSource'))

        self.assertEqual(
            '"new-york.power.usage" 42422.0 1493773500 source="localhost:5050"'
            ' "datacenter"="dc1"\n',
            metric_to_line_data('new-york.power.usage', 42422, 1493773500,
                                'localhost:5050', {'datacenter': 'dc1'},
                                'defaultSource'))

        # null timestamp
        self.assertEqual(
            '"new-york.power.usage" 42422.0 source="localhost" '
            '"datacenter"="dc1"\n',
            metric_to_line_data('new-york.power.usage', 42422, None,
                                'localhost', {'datacenter': 'dc1'},
                                'defaultSource'))

        # null tags
        self.assertEqual(
            '"new-york.power.usage" 42422.0 1493773500 source="localhost"\n',
            metric_to_line_data('new-york.power.usage', 42422, 1493773500,
                                'localhost', None, 'defaultSource'))

        # null tags and null timestamp
        self.assertEqual(
            '"new-york.power.usage" 42422.0 source="localhost"\n',
            metric_to_line_data('new-york.power.usage', 42422, None,
                                'localhost', None, 'defaultSource'))

    def test_histogram_to_line_data(self):
        """Test wavefront_sdk.common.utils.histogram_to_line_data()."""
        self.assertEqual(
            '!M 1493773500 #20 30.0 #10 5.1 "request.latency" '
            'source="appServer1" "region"="us-west"\n',
            histogram_to_line_data('request.latency', [(30.0, 20), (5.1, 10)],
                                   {histogram_granularity.MINUTE}, 1493773500,
                                   'appServer1', {'region': 'us-west'},
                                   'defaultSource'))

        self.assertEqual(
            '!M 1493773500 #20 30.0 #10 5.1 "request.latency" '
            'source="127.0.0.1:5050" "region"="us-west"\n',
            histogram_to_line_data('request.latency', [(30.0, 20), (5.1, 10)],
                                   {histogram_granularity.MINUTE}, 1493773500,
                                   '127.0.0.1:5050', {'region': 'us-west'},
                                   'defaultSource'))

        # null timestamp
        self.assertEqual(
            '!M #20 30.0 #10 5.1 "request.latency" source="appServer1" '
            '"region"="us-west"\n',
            histogram_to_line_data('request.latency', [(30.0, 20), (5.1, 10)],
                                   {histogram_granularity.MINUTE},
                                   None, 'appServer1', {'region': 'us-west'},
                                   'defaultSource'))

        # null tags
        self.assertEqual(
            '!M 1493773500 #20 30.0 #10 5.1 "request.latency" '
            'source="appServer1"\n',
            histogram_to_line_data('request.latency', [(30.0, 20), (5.1, 10)],
                                   {histogram_granularity.MINUTE}, 1493773500,
                                   'appServer1', None, 'defaultSource'))

        # empty centroids
        with self.assertRaises(ValueError):
            histogram_to_line_data('request.latency', [],
                                   {histogram_granularity.MINUTE},
                                   1493773500, 'appServer1', None,
                                   'defaultSource')

        # no histogram granularity specified
        with self.assertRaises(ValueError):
            histogram_to_line_data('request.latency', [(30.0, 20), (5.1, 10)],
                                   {}, 1493773500, 'appServer1', None,
                                   'defaultSource')

        # multiple granularities
        self.assertEqual(
            sorted(['!M 1493773500 #20 30.0 #10 5.1 "request.latency" '
                    'source="appServer1" "region"="us-west"',
                    '!H 1493773500 #20 30.0 #10 5.1 "request.latency" '
                    'source="appServer1" "region"="us-west"',
                    '!D 1493773500 #20 30.0 #10 5.1 "request.latency" '
                    'source="appServer1" "region"="us-west"']),
            sorted(
                histogram_to_line_data(
                    'request.latency', [(30.0, 20), (5.1, 10)],
                    {histogram_granularity.MINUTE, histogram_granularity.HOUR,
                     histogram_granularity.DAY},
                    1493773500, 'appServer1', {'region': 'us-west'},
                    'defaultSource'
                ).splitlines()))

    def test_tracing_span_to_line_data(self):
        """Test wavefront_sdk.common.utils.tracing_span_to_line_data()."""
        self.assertEqual(
            '"getAllUsers" source="localhost" '
            'traceId=7b3bf470-9456-11e8-9eb6-529269fb1459 '
            'spanId=0313bafe-9457-11e8-9eb6-529269fb1459 '
            'parent=2f64e538-9457-11e8-9eb6-529269fb1459 '
            'followsFrom=5f64e538-9457-11e8-9eb6-529269fb1459 '
            '"application"="Wavefront" '
            '"http.method"="GET" 1493773500 343500\n',
            tracing_span_to_line_data(
                'getAllUsers', 1493773500, 343500, 'localhost',
                uuid.UUID('7b3bf470-9456-11e8-9eb6-529269fb1459'),
                uuid.UUID('0313bafe-9457-11e8-9eb6-529269fb1459'),
                [uuid.UUID('2f64e538-9457-11e8-9eb6-529269fb1459')],
                [uuid.UUID('5f64e538-9457-11e8-9eb6-529269fb1459')],
                [('application', 'Wavefront'), ('http.method', 'GET')],
                None, 'defaultSource'))

        self.assertEqual(
            '"getAllUsers" source="localhost:5050" '
            'traceId=7b3bf470-9456-11e8-9eb6-529269fb1459 '
            'spanId=0313bafe-9457-11e8-9eb6-529269fb1459 '
            'parent=2f64e538-9457-11e8-9eb6-529269fb1459 '
            'followsFrom=5f64e538-9457-11e8-9eb6-529269fb1459 '
            '"application"="Wavefront" '
            '"http.method"="GET" 1493773500 343500\n',
            tracing_span_to_line_data(
                'getAllUsers', 1493773500, 343500, 'localhost:5050',
                uuid.UUID('7b3bf470-9456-11e8-9eb6-529269fb1459'),
                uuid.UUID('0313bafe-9457-11e8-9eb6-529269fb1459'),
                [uuid.UUID('2f64e538-9457-11e8-9eb6-529269fb1459')],
                [uuid.UUID('5f64e538-9457-11e8-9eb6-529269fb1459')],
                [('application', 'Wavefront'),
                 ('http.method', 'GET')], None, 'defaultSource'))

        # null followsFrom
        self.assertEqual(
            '"getAllUsers" source="localhost" '
            'traceId=7b3bf470-9456-11e8-9eb6-529269fb1459 '
            'spanId=0313bafe-9457-11e8-9eb6-529269fb1459 '
            'parent=2f64e538-9457-11e8-9eb6-529269fb1459 '
            '"application"="Wavefront" '
            '"http.method"="GET" 1493773500 343500\n',
            tracing_span_to_line_data(
                'getAllUsers', 1493773500, 343500, 'localhost',
                uuid.UUID('7b3bf470-9456-11e8-9eb6-529269fb1459'),
                uuid.UUID('0313bafe-9457-11e8-9eb6-529269fb1459'),
                [uuid.UUID('2f64e538-9457-11e8-9eb6-529269fb1459')],
                None, [('application', 'Wavefront'), ('http.method', 'GET')],
                None, 'defaultSource'))

        # root span
        self.assertEqual(
            '"getAllUsers" source="localhost" '
            'traceId=7b3bf470-9456-11e8-9eb6-529269fb1459 '
            'spanId=0313bafe-9457-11e8-9eb6-529269fb1459 '
            '"application"="Wavefront" '
            '"http.method"="GET" 1493773500 343500\n',
            tracing_span_to_line_data(
                'getAllUsers', 1493773500, 343500, 'localhost',
                uuid.UUID('7b3bf470-9456-11e8-9eb6-529269fb1459'),
                uuid.UUID('0313bafe-9457-11e8-9eb6-529269fb1459'), None, None,
                [('application', 'Wavefront'), ('http.method', 'GET')],
                None, 'defaultSource'))

        # duplicate tags
        self.assertEqual(
            '"getAllUsers" source="localhost" '
            'traceId=7b3bf470-9456-11e8-9eb6-529269fb1459 '
            'spanId=0313bafe-9457-11e8-9eb6-529269fb1459 '
            '"application"="Wavefront" '
            '"http.method"="GET" 1493773500 343500\n',
            tracing_span_to_line_data(
                'getAllUsers', 1493773500, 343500, 'localhost',
                uuid.UUID('7b3bf470-9456-11e8-9eb6-529269fb1459'),
                uuid.UUID('0313bafe-9457-11e8-9eb6-529269fb1459'), None, None,
                [('application', 'Wavefront'), ('http.method', 'GET'),
                 ('application', 'Wavefront')],
                None, 'defaultSource'))

        # null tags
        self.assertEqual(
            '"getAllUsers" source="localhost" '
            'traceId=7b3bf470-9456-11e8-9eb6-529269fb1459 '
            'spanId=0313bafe-9457-11e8-9eb6-529269fb1459 '
            '1493773500 343500\n',
            tracing_span_to_line_data(
                'getAllUsers', 1493773500, 343500, 'localhost',
                uuid.UUID('7b3bf470-9456-11e8-9eb6-529269fb1459'),
                uuid.UUID('0313bafe-9457-11e8-9eb6-529269fb1459'),
                None, None, None, None, 'defaultSource'))


if __name__ == '__main__':
    # run 'python -m unittest discover' from top-level to run tests
    unittest.main()



# -*- coding: utf-8 -*-
"""Wavefront SDK Metrics Registry.

@author Hao Song (songhao@vmware.com)
"""
import logging
import threading
import time

from wavefront_sdk.common.metrics import counter, gauge

LOGGER = logging.getLogger('wavefront_sdk.WavefrontSdkMetricsRegistry')


# pylint: disable=too-many-instance-attributes,E0012,R0205
class WavefrontSdkMetricsRegistry(object):
    """Wavefront SDK Metrics Registry."""

    # pylint: disable=too-many-arguments
    def __init__(self, wf_metric_sender, source=None, tags=None, prefix=None,
                 reporting_interval_secs=60):
        """Construct Wavefront SDK Metrics Registry."""
        self.wf_metric_sender = wf_metric_sender
        self.source = source
        self.tags = tags
        self.prefix = '' if not prefix else prefix + '.'
        self.reporting_interval_secs = reporting_interval_secs
        self.metrics = {}
        self._closed = False
        self._schedule_lock = threading.Lock()
        self._timer = None
        if wf_metric_sender:
            self._schedule_timer()

    def _schedule_timer(self):
        if not self._closed:
            self._timer = threading.Timer(self.reporting_interval_secs,
                                          self._run)
            self._timer.daemon = True
            self._timer.start()

    # pylint: disable=broad-except
    def _report(self, timeout_secs=None):
        timestamp = time.time()
        for key, val in self.metrics.items():
            if timeout_secs and time.time() - timestamp > timeout_secs:
                break
            name = self.prefix + key
            try:
                if isinstance(val, gauge.WavefrontSdkGauge):
                    gauge_value = val.get_value()
                    if gauge_value:
                        self.wf_metric_sender.send_metric(
                            name, gauge_value, timestamp, self.source,
                            self.tags)
                elif isinstance(val, counter.WavefrontSdkCounter):
                    self.wf_metric_sender.send_metric(
                        name + '.count', val.count(), timestamp,
                        self.source, self.tags)
            except Exception:
                LOGGER.warning('Unable to send internal SDK metric.')

    def _run(self):
        try:
            self._report()
        finally:
            with self._schedule_lock:
                if not self._closed:
                    self._schedule_timer()

    def close(self, timeout_secs=None):
        """Close Wavefront SDK Metrics Registry."""
        try:
            if self.wf_metric_sender:
                self._report(timeout_secs)
        finally:
            with self._schedule_lock:
                self._closed = True
                if self._timer is not None:
                    self._timer.cancel()

    def new_counter(self, name):
        """Get or create a counter from the registry."""
        return self._get_or_add(name, counter.WavefrontSdkCounter())

    def new_gauge(self, name, supplier):
        """Get or create a gauge from the registry."""
        return self._get_or_add(name, gauge.WavefrontSdkGauge(supplier))

    def _get_or_add(self, name, metric):
        existing_metric = self.metrics.get(name)
        if existing_metric:
            return existing_metric
        self.metrics.update({name: metric})
        return metric



# -*- coding: utf-8 -*-
"""Wavefront Python SDK.

This library provides support for sending metrics, histograms and opentracing
spans to Wavefront via proxy or direct ingestion.

@author Hao Song (songhao@vmware.com)
"""

import pkg_resources

from .direct import WavefrontDirectClient
from .proxy import WavefrontProxyClient


__all__ = ['WavefrontDirectClient', 'WavefrontProxyClient']

__version__ = None

try:
    __version__ = pkg_resources.get_distribution(
        'wavefront-sdk-python'
    ).version
except pkg_resources.DistributionNotFound:
    # __version__ is only available when distribution is installed.
    pass



"""Unit Tests for Wavefront Python SDK - Histogram Impl.

@author Hao Song (songhao@vmware.com)
"""

import threading
import time
import unittest

from wavefront_sdk.common import utils
from wavefront_sdk.entities import WavefrontHistogramImpl


class TestHistogramImpl(unittest.TestCase):
    """Unit Tests of WavefrontHistogramImpl."""

    @classmethod
    def setUpClass(cls):
        """Initialize for tests."""
        # Delta for error
        cls._DELTA = 1e-1

        # Atomic clock
        cls._clock = utils.AtomicCounter(time.time() * 1000)

        # WavefrontHistogramImpl with values that are powers of 10
        cls._pow_10 = cls.create_pow_10_histogram(cls._clock.get)

        # WavefrontHistogramImpl with a value for each integer from 1 to 100
        cls._inc_100 = WavefrontHistogramImpl(cls._clock.get)
        for i in range(1, 101):
            cls._inc_100.update(i)

        # WavefrontHistogramImpl with a value for each integer from 1 to 1000
        cls._inc_1000 = WavefrontHistogramImpl(cls._clock.get)
        for i in range(1, 1001):
            cls._inc_1000.update(i)

        # Empty Wavefront Histogram
        cls._empty_hist = WavefrontHistogramImpl(cls._clock.get)

        # Simulate that 1 min has passed so that values prior to the
        # current min are ready to be read
        cls._clock.increment(60000 + 1)

    @staticmethod
    def create_pow_10_histogram(clock_millis):
        """Create WavefrontHistogramImpl with values that are powers of 10."""
        w_h = WavefrontHistogramImpl(clock_millis)
        w_h.update(0.1)
        w_h.update(1.0)
        w_h.update(1e1)
        w_h.update(1e1)
        w_h.update(1e2)
        w_h.update(1e3)
        w_h.update(1e4)
        w_h.update(1e4)
        w_h.update(1e5)
        return w_h

    @staticmethod
    def distribution_to_map(distributions):
        """Return Distributions in map format.

        @return: Distributions in map format.
        @rtype: dict
        """
        dist_map = {}
        for distribution in distributions:
            for centroid in distribution.centroids:
                dist_map.update(
                    {centroid[0]: dist_map.get(centroid[0], 0) + centroid[1]})
        return dist_map

    @staticmethod
    def thread_bulk_update(wavefront_histogram, means, counts):
        """Act as a helper func for multi-thread bulk updating."""
        while True:
            wavefront_histogram.bulk_update(means, counts)
            time.sleep(60)

    def test_distribution(self):
        """Test distribution."""
        w_h = self.create_pow_10_histogram(self._clock.get)
        self._clock.increment(60000 + 1)

        distributions = w_h.flush_distributions()
        dist_map = self.distribution_to_map(distributions)
        self.assertEqual(7, len(dist_map))
        self.assertTrue(dist_map.get(0.1) == 1)
        self.assertTrue(dist_map.get(1.0) == 1)
        self.assertTrue(dist_map.get(1e1) == 2)
        self.assertTrue(dist_map.get(1e2) == 1)
        self.assertTrue(dist_map.get(1e3) == 1)
        self.assertTrue(dist_map.get(1e4) == 2)
        self.assertTrue(dist_map.get(1e5) == 1)

        self.assertEqual(0, w_h.get_count())
        self.assertAlmostEqual(None, w_h.get_max(), delta=self._DELTA)
        self.assertAlmostEqual(None, w_h.get_min(), delta=self._DELTA)
        self.assertAlmostEqual(None, w_h.get_mean(), delta=self._DELTA)
        self.assertAlmostEqual(0, w_h.get_sum(), delta=self._DELTA)

        snapshot = w_h.get_snapshot()
        self.assertEqual(0, snapshot.get_count())
        self.assertAlmostEqual(None, snapshot.get_max(), delta=self._DELTA)
        self.assertAlmostEqual(None, snapshot.get_min(), delta=self._DELTA)
        self.assertAlmostEqual(0, snapshot.get_mean(), delta=self._DELTA)
        self.assertAlmostEqual(0, snapshot.get_sum(), delta=self._DELTA)
        self.assertAlmostEqual(None, snapshot.get_value(0.5),
                               delta=self._DELTA)

    def test_bulk_update(self):
        """Test bulk update."""
        w_h = WavefrontHistogramImpl(self._clock.get)
        w_h.bulk_update([24.2, 84.35, 1002.0], [80, 1, 9])
        self._clock.increment(60000 + 1)
        distributions = w_h.flush_distributions()
        dist_map = self.distribution_to_map(distributions)

        self.assertEqual(3, len(dist_map))
        self.assertTrue(dist_map.get(24.2) == 80)
        self.assertTrue(dist_map.get(84.35) == 1)
        self.assertTrue(dist_map.get(1002.0) == 9)

    def test_count(self):
        """Test get count of distribution."""
        self.assertEqual(9, self._pow_10.get_count())
        self.assertEqual(9, self._pow_10.get_snapshot().get_count())

    def test_max(self):
        """Test get max of distribution."""
        self.assertAlmostEqual(1e5, self._pow_10.get_max(), delta=self._DELTA)
        self.assertAlmostEqual(1e5, self._pow_10.get_snapshot().get_max(),
                               delta=self._DELTA)

    def test_min(self):
        """Test get min of distribution."""
        self.assertAlmostEqual(1.0, self._inc_100.get_min(), delta=self._DELTA)
        self.assertAlmostEqual(1.0, self._inc_100.get_snapshot().get_min(),
                               delta=self._DELTA)

    def test_mean(self):
        """Test get mean of distribution."""
        self.assertAlmostEqual(13457.9, self._pow_10.get_mean(),
                               delta=self._DELTA)
        self.assertAlmostEqual(13457.9, self._pow_10.get_snapshot().get_mean(),
                               delta=self._DELTA)

    def test_sum(self):
        """Test get sum of distribution."""
        self.assertAlmostEqual(121121.1, self._pow_10.get_sum(),
                               delta=self._DELTA)
        self.assertAlmostEqual(121121.1, self._pow_10.get_snapshot().get_sum(),
                               delta=self._DELTA)

    def test_std_dev(self):
        """Test get std dev of distribution."""
        self.assertAlmostEqual(30859.85, self._pow_10.std_dev(),
                               delta=self._DELTA)
        self.assertAlmostEqual(28.87, self._inc_100.std_dev(),
                               delta=self._DELTA)
        self.assertAlmostEqual(288.67, self._inc_1000.std_dev(),
                               delta=self._DELTA)
        self.assertEqual(0.0, self._empty_hist.std_dev())

    def test_size(self):
        """Test get size of distribution."""
        self.assertEqual(9, self._pow_10.get_snapshot().get_size())
        self.assertEqual(100, self._inc_100.get_snapshot().get_size())
        self.assertEqual(1000, self._inc_1000.get_snapshot().get_size())

    def test_multi_thread(self):
        """Test of multi-threading case."""
        w_h = WavefrontHistogramImpl(self._clock.get)
        w_h.bulk_update([21.2, 82.35, 1042.0], [70, 2, 6])
        thread_1 = threading.Thread(
            target=self.thread_bulk_update,
            args=(w_h, [24.2, 84.35, 1002.0], [80, 1, 9]))
        thread_2 = threading.Thread(
            target=self.thread_bulk_update,
            args=(w_h, [21.2, 84.35, 1052.0], [60, 12, 8]))
        thread_1.setDaemon(True)
        thread_2.setDaemon(True)
        thread_1.start()
        thread_2.start()
        time.sleep(1)
        self._clock.increment(60000 + 1)
        distributions = w_h.flush_distributions()
        dist_map = self.distribution_to_map(distributions)
        thread_1.join(2)
        thread_2.join(2)

        self.assertEqual(7, len(dist_map))
        self.assertTrue(dist_map.get(21.2) == 130)
        self.assertTrue(dist_map.get(24.2) == 80)
        self.assertTrue(dist_map.get(82.35) == 2)
        self.assertTrue(dist_map.get(84.35) == 13)
        self.assertTrue(dist_map.get(1002.0) == 9)
        self.assertTrue(dist_map.get(1042.0) == 6)
        self.assertTrue(dist_map.get(1052.0) == 8)


if __name__ == '__main__':
    # run 'python -m unittest discover' from top-level to run tests
    unittest.main()



# -*- coding: utf-8 -*-
"""Wavefront SDK Metric.

@author Hao Song (songhao@vmware.com)
"""


# pylint: disable=too-few-public-methods,E0012,R0205
class WavefrontSdkMetric(object):
    """Wavefront SDK Metric."""



"""All Python-sdk constants.

@author Hao Song (songhao@vmware.com).
"""

# Use this format to send metric data to Wavefront.
WAVEFRONT_METRIC_FORMAT = 'wavefront'

# Use this format to send histogram data to Wavefront.
WAVEFRONT_HISTOGRAM_FORMAT = 'histogram'

# Use this format to send tracing data to Wavefront.
WAVEFRONT_TRACING_SPAN_FORMAT = 'trace'

# GREEK LETTER DELTA.
DELTA_PREFIX = u'\u2206'

# GREEK CAPITAL LETTER DELTA.
DELTA_PREFIX_2 = u'\u0394'

# Heartbeat metric.
HEART_BEAT_METRIC = '~component.heartbeat'

# Internal source used for internal and aggregated metrics.
WAVEFRONT_PROVIDED_SOURCE = 'wavefront-provided'

# Null value emitted for optional undefined tags.
NULL_TAG_VAL = 'none'

# Key for defining a source.
SOURCE_KEY = 'source'

# Tag key for defining an application.
APPLICATION_TAG_KEY = 'application'

# Tag key for defining a cluster.
CLUSTER_TAG_KEY = 'cluster'

# Tag key for defining a shard.
SHARD_TAG_KEY = 'shard'

# Tag key  for defining a service.
SERVICE_TAG_KEY = 'service'

# Tag key for defining a component.
COMPONENT_TAG_KEY = 'component'

# Heart beat interval.
HEART_BEAT_INTERVAL = 10

# Name prefix for internal diagnostic metrics for Wavefront SDKs.
SDK_METRIC_PREFIX = '~sdk.python'

# Tag key for span logs.
SPAN_LOG_KEY = "_spanLogs"

# Distribution name for sdk
WAVEFRONT_SDK_PYTHON = 'wavefront-sdk-python'



# -*- coding: utf-8 -*-
"""Events Entities.

@author Yogesh Prasad Kurmi (ykurmi@vmware.com)
"""


# pylint: disable=E0012,R0205
class WavefrontEventSender(object):
    """Interface of Event Sender for both clients."""

    # pylint: disable=too-many-arguments
    def send_event(self, name, start_time, end_time, source,
                   tags, annotations):
        """Send Event Data.

        Wavefront Event Data format
        {"name": <Event Name>, "annotations": <Annotations>,
         "hosts": <Host List>,"startTime": <Start Time>,
          "endTime": <End Time>, "tags": <Tags>}
        Example: {"name": event_via_direct_ingestion, "annotations": {
        "severity": "severe", "type": "backup", "details": "broker backup"},
         "hosts": ["localhost"], "startTime": 1590678089,
          "endTime": 1590679089, "tags": ["env:", "test"]}

        @param name: Event Name
        @type name: str
        @param start_time: Event Start Time
        @type start_time: long
        @param end_time: Event End Time
        @type end_time: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: list[str]
        @param annotations: Annotations
        @type annotations: dict
        """
        raise NotImplementedError

    def send_event_now(self, events):
        """Send a list of events immediately.

        @param events: List of string events data
        @type events: list[str]
        """
        raise NotImplementedError



"""Service that periodically reports component heartbeats to Wavefront.

@author Hao Song (songhao@vmware.com).
"""

import logging
import threading
import time

from wavefront_sdk.common.utils import HashableDict

from .constants import APPLICATION_TAG_KEY
from .constants import CLUSTER_TAG_KEY
from .constants import COMPONENT_TAG_KEY
from .constants import HEART_BEAT_METRIC
from .constants import NULL_TAG_VAL
from .constants import SERVICE_TAG_KEY
from .constants import SHARD_TAG_KEY


# pylint: disable=E0012,R0205
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-instance-attributes
class HeartbeaterService(object):
    """Service that periodically reports component heartbeats to Wavefront."""

    # pylint: disable=too-many-arguments
    def __init__(self, wavefront_client, application_tags, components, source):
        """Construct HeartbeaterService.

        @param wavefront_client: Wavefront Proxy or Direct Ingestion client.
        @param application_tags: ApplicationTags.
        @param components: List of str indicates Components.
        @param source: Source.
        """
        self.wavefront_client = wavefront_client
        self.application_tags = application_tags
        self.source = source
        self.reporting_interval_seconds = 60 * 5
        self.heartbeat_metric_tags_list = []
        self.custom_tags_set = set()
        if isinstance(components, str):
            components = [components]
        for component in components:
            metric_tags = {
                APPLICATION_TAG_KEY: application_tags.application,
                CLUSTER_TAG_KEY: application_tags.cluster or NULL_TAG_VAL,
                SERVICE_TAG_KEY: application_tags.service,
                SHARD_TAG_KEY: application_tags.shard or NULL_TAG_VAL,
                COMPONENT_TAG_KEY: component
            }
            if application_tags.custom_tags:
                metric_tags.update(dict(application_tags.custom_tags))
            self.heartbeat_metric_tags_list.append(metric_tags)
        self._closed = False
        self._timer = None
        self._run()

    def report_custom_tags(self, custom_tags):
        """
        Append custom tags for heartbeat reporting.

        @param custom_tags: dict of custom tags.
        """
        self.custom_tags_set.add(HashableDict(custom_tags))

    def _schedule_timer(self):
        self._timer = threading.Timer(self.reporting_interval_seconds,
                                      self._run)
        self._timer.daemon = True
        self._timer.start()

    def _run(self):
        try:
            self._report()
        finally:
            if not self._closed:
                self._schedule_timer()

    def _report(self):
        try:
            for heartbeat in self.heartbeat_metric_tags_list:
                self.wavefront_client.send_metric(HEART_BEAT_METRIC, 1.0,
                                                  time.time(), self.source,
                                                  heartbeat)
            while self.custom_tags_set:
                self.wavefront_client.send_metric(HEART_BEAT_METRIC, 1.0,
                                                  time.time(), self.source,
                                                  self.custom_tags_set.pop())

        # pylint: disable=broad-except,fixme
        # TODO: Please make sure we catch more specific exception here.
        except Exception:
            logging.warning('Can not report %s to wavefront',
                            HEART_BEAT_METRIC)

    def close(self):
        """Cancel the timer."""
        self._closed = True
        if self._timer is not None:
            self._timer.cancel()



# -*- coding: utf-8 -*-
"""Connection Handler class for sending data to a Wavefront proxy.

@author: Hao Song (songhao@vmware.com)
"""

from __future__ import absolute_import

import socket

from . import connection_handler


# pylint: disable=too-many-instance-attributes
class ProxyConnectionHandler(connection_handler.ConnectionHandler):
    """Connection Handler.

    For sending data to a Wavefront proxy listening on a given port.
    """

    # pylint: disable=too-many-arguments
    def __init__(self, address, port, wavefront_sdk_metrics_registry,
                 entity_prefix=None, timeout=None):
        """Construct ProxyConnectionHandler.

        @param address: Proxy Address
        @param port: Proxy Port
        """
        super().__init__()
        self._address = address
        self._port = int(port)
        self.entity_prefix = '' if not entity_prefix else entity_prefix + ''
        self.timeout = timeout
        self.wf_metrics_registry = wavefront_sdk_metrics_registry
        self._write_successes = self.wf_metrics_registry.new_counter(
            self.entity_prefix + 'write.success')
        self._write_errors = self.wf_metrics_registry.new_counter(
            self.entity_prefix + 'write.errors')
        self._reconnecting_socket = None

    def connect(self):
        """Initialize socket and connect to given address:port."""
        self._reconnecting_socket = socket.socket(socket.AF_INET,
                                                  socket.SOCK_STREAM)
        self._reconnecting_socket.settimeout(self.timeout)
        self._reconnecting_socket.connect((self._address, self._port))

    def close(self):
        """Close socket if it's open now."""
        if self._reconnecting_socket:
            self._reconnecting_socket.close()

    def send_data(self, line_data, reconnect=True):
        """Send data via proxy.

        @param line_data: Data to be sent
        @param reconnect: If it's the second time trying to send data
        """
        try:
            if not self._reconnecting_socket:
                self.connect()
            self._reconnecting_socket.sendall(line_data.encode('utf-8'))
            self._write_successes.inc()
        except socket.error as error:
            if reconnect:
                self._reconnecting_socket = None
                # Try to resend
                self.send_data(line_data, reconnect=False)
            else:
                # Second time trying failed
                self._write_errors.inc()
                raise error



"""Unit Tests for Wavefront SDK Metrics Registry.

@author Hao Song (songhao@vmware.com)
"""

import unittest

try:
    import queue
except ImportError:
    import Queue as queue  # noqa

from wavefront_sdk.common.metrics.registry import WavefrontSdkMetricsRegistry
from wavefront_sdk.direct import remaining_capacity_getter


class TestUtils(unittest.TestCase):
    """Test Functions of wavefront_sdk.common.utils."""

    def test_gauge(self):
        """Test WavefrontSdkGauge of WavefrontSdkMetricsRegistry."""
        registry = WavefrontSdkMetricsRegistry(None)
        capacity = 5
        buffer = queue.Queue(capacity)
        size = registry.new_gauge('gauge', buffer.qsize)
        remaining_capacity = registry.new_gauge(
            'remaining', remaining_capacity_getter(buffer))
        self.assertEqual(size.get_value(), 0)
        self.assertEqual(remaining_capacity.get_value(), capacity)
        buffer.put(0)
        self.assertEqual(size.get_value(), 1)
        self.assertEqual(registry.new_gauge('gauge', None).get_value(), 1)
        self.assertEqual(remaining_capacity.get_value(), capacity - 1)

    def test_counter(self):
        """Test WavefrontSdkCounter of WavefrontSdkMetricsRegistry."""
        registry = WavefrontSdkMetricsRegistry(None)
        counter = registry.new_counter('counter')
        self.assertEqual(counter.count(), 0)
        counter.inc()
        self.assertEqual(counter.count(), 1)
        counter.inc(2)
        self.assertEqual(counter.count(), 3)
        self.assertEqual(registry.new_counter('counter').count(), 3)
        self.assertEqual(registry.new_counter('counter_2').count(), 0)


if __name__ == '__main__':
    # run 'python -m unittest discover' from top-level to run tests
    unittest.main()



# -*- coding: utf-8 -*-
"""Wavefront SDK Counter.

@author Hao Song (songhao@vmware.com)
"""
import threading

from wavefront_sdk.common.metrics import metrics


class WavefrontSdkCounter(metrics.WavefrontSdkMetric):
    """Wavefront SDK Counter."""

    def __init__(self):
        """Construct Wavefront SDK Counter."""
        self._lock = threading.Lock()
        self._count = 0

    def inc(self, val=1):
        """Increase the value of the counter."""
        with self._lock:
            self._count += val

    def count(self):
        """Get the value of the counter."""
        with self._lock:
            return self._count

    def clear(self):
        """Reset the counter."""
        with self._lock:
            self._count = 0



"""Metadata about your application represented as tags in Wavefront.

@author: Hao Song (songhao@vmware.com)
"""
import os
import re

from .constants import APPLICATION_TAG_KEY
from .constants import CLUSTER_TAG_KEY
from .constants import NULL_TAG_VAL
from .constants import SERVICE_TAG_KEY
from .constants import SHARD_TAG_KEY


# pylint: disable=E0012,R0205
class ApplicationTags(object):
    """Metadata about your application represented as tags in Wavefront."""

    # pylint: disable=too-many-arguments
    def __init__(self, application, service, cluster=None, shard=None,
                 custom_tags=None):
        """Construct ApplicationTags.

        @param application: Application Name
        @param service: Service Name
        @param cluster: Cluster Name
        @param shard: Shard Name
        @param custom_tags: List of pairs of custom tags
        """
        if not application:
            raise AttributeError('Missing "application" parameter in '
                                 'ApplicationTags!')
        if not service:
            raise AttributeError('Missing "service" parameter in '
                                 'ApplicationTags!')
        self._application = application
        self._service = service
        self._cluster = cluster
        self._shard = shard
        self._custom_tags = []
        if custom_tags:
            self._custom_tags.extend(custom_tags)

    @property
    def application(self):
        """Get Application Name."""
        return self._application

    @property
    def service(self):
        """Get Service Name."""
        return self._service

    @property
    def cluster(self):
        """Get Cluster Name."""
        return self._cluster

    @property
    def shard(self):
        """Get Shard Name."""
        return self._shard

    @property
    def custom_tags(self):
        """Get Custom Tags."""
        return self._custom_tags

    def get_as_list(self):
        """Get all tags as a list."""
        tags = [(APPLICATION_TAG_KEY, self.application),
                (SERVICE_TAG_KEY, self.service),
                (CLUSTER_TAG_KEY, self.cluster or NULL_TAG_VAL),
                (SHARD_TAG_KEY, self.shard or NULL_TAG_VAL)]
        if self.custom_tags:
            tags.extend(self.custom_tags)
        return tags

    def add_custom_tags_from_env(self, pattern):
        """Set custom tags from environment variables that match the regex pattern.

        @param pattern: Regex pattern
        """
        for key in os.environ:
            if re.match(pattern, key, re.IGNORECASE):
                value = os.environ[key]
                if value:
                    self._custom_tags.append((key, value))

    def add_custom_tag_from_env(self, tag_key, var_key):
        """Set a custom tag from the given environment variable.

        @param tag_key: Key of the custom tag
        @param var_name: Key of the environment variable
        """
        if var_key in os.environ:
            value = os.environ[var_key]
            if value:
                self._custom_tags.append((tag_key, value))
        else:
            raise Exception("var_key is an invalid environment variable.")



# -*- coding: utf-8 -*-
"""Entities for SDK.

@author Hao Song (songhao@vmware.com)
"""

from .event.sender import WavefrontEventSender
from .histogram import histogram_granularity
from .histogram.histogram_impl import WavefrontHistogramImpl
from .histogram.sender import WavefrontHistogramSender
from .metrics.sender import WavefrontMetricSender
from .tracing.sender import WavefrontTracingSpanSender


__all__ = ['histogram_granularity',
           'WavefrontHistogramImpl',
           'WavefrontHistogramSender',
           'WavefrontMetricSender',
           'WavefrontEventSender',
           'WavefrontTracingSpanSender']



# -*- coding: utf-8 -*-
"""Tracing Span Sender Interface for both Clients.

@author Hao Song (songhao@vmware.com)`
"""


# pylint: disable=E0012,R0205
class WavefrontTracingSpanSender(object):
    """Tracing Span Sender Interface for both Clients."""

    # pylint: disable=too-many-arguments
    def send_span(self, name, start_millis, duration_millis, source, trace_id,
                  span_id, parents, follows_from, tags, span_logs):
        """Send span data via proxy.

        Wavefront Tracing Span Data format
        <tracingSpanName> source=<source> [pointTags] <start_millis>
        <duration_milli_seconds>
        Example: "getAllUsers source=localhost
                  traceId=7b3bf470-9456-11e8-9eb6-529269fb1459
                  spanId=0313bafe-9457-11e8-9eb6-529269fb1459
                  parent=2f64e538-9457-11e8-9eb6-529269fb1459
                  application=Wavefront http.method=GET
                  1533531013 343500"

        @param name: Span Name
        @type name: str
        @param start_millis: Start time
        @type start_millis: long
        @param duration_millis: Duration time
        @type duration_millis: long
        @param source: Source
        @type source: str
        @param trace_id: Trace ID
        @type trace_id: UUID
        @param span_id: Span ID
        @type span_id: UUID
        @param parents: Parents Span ID
        @type parents: List of UUID
        @param follows_from: Follows Span ID
        @type follows_from: List of UUID
        @param tags: Tags
        @type tags: list
        @param span_logs: Span Log
        """
        raise NotImplementedError

    def send_formatted_span(self, span):
        """Send a formatted span immediately.

        @param span: Formatted span
        @type: str
        """
        self.send_span_now([span])

    def send_span_now(self, spans):
        """Send a list of spans immediately.

        Have to construct the data manually by calling
        common.utils.metric_to_line_data()

        @param spans: List of string spans data
        @type spans: list[str]
        """
        raise NotImplementedError

    def send_span_log_now(self, span_logs):
        """Send a list of span logs immediately.

        Have to construct the data manually by calling
        common.utils.span_log_to_line_data()

        @param span_logs: List of string span logs data
        @type span_logs: list[str]
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""Provide utils functions.

@author Hao Song (songhao@vmware.com)
"""

from .application_tags import ApplicationTags
from .heartbeater_service import HeartbeaterService
from .utils import event_to_json
from .utils import event_to_line_data
from .utils import histogram_to_line_data
from .utils import metric_to_line_data
from .utils import tracing_span_to_line_data

__all__ = ['histogram_to_line_data',
           'metric_to_line_data',
           'tracing_span_to_line_data',
           'event_to_json',
           'event_to_line_data',
           'ApplicationTags',
           'HeartbeaterService']



# -*- coding: utf-8 -*-
"""
Utils module contains useful function for preparing and processing data.

@author: Hao Song (songhao@vmware.com)
"""

import gzip
import io
import json
import logging
import re
import threading

import pkg_resources

from wavefront_sdk.common.constants import SPAN_LOG_KEY

LOGGER = logging.getLogger('wavefront_sdk.utils')


# pylint: disable=E0012,R0205
class AtomicCounter(object):
    """An atomic, thread-safe incrementing counter."""

    def __init__(self, initial=0):
        """Construct Atomic Counter.

        @param initial: Initial value of the counter
        """
        self.value = initial
        self._lock = threading.Lock()

    def increment(self, num=1):
        """Increment atomic counter value.

        @param num: Num to be increased, 1 by default
        @return: Current value after increment
        """
        with self._lock:
            self.value += num
            return self.value

    def get(self):
        """Get current atomic counter value.

        @return: Current atomic counter value.
        @rtype: float or int
        """
        return self.value


class HashableDict(dict):
    """Hashable Dict."""

    def __hash__(self):
        """Hash of the dict."""
        return hash(tuple(sorted(self.items())))


def chunks(data_list, batch_size):
    """Split list of data into chunks with fixed batch size.

    @param data_list: List of data
    @param batch_size: Batch size of each chunk
    @return: Return a lazy generator object for iteration
    """
    for i in range(0, len(data_list), batch_size):
        yield data_list[i:i + batch_size]


def gzip_compress(data, compresslevel=9):
    """Compress data using GZIP.

    @param data: Data to compress
    @param compresslevel: Compress Level
    @return: Compressed data
    """
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, mode='wb',
                       compresslevel=compresslevel) as gzip_file:
        gzip_file.write(data)
    return buf.getvalue()


def sanitize(string):
    """Sanitize a string with quotes.

    @param string: Input string
    @return: Sanitized string
    """
    return sanitize_internal(string, True)


def sanitize_without_quotes(string):
    """Sanitize a string without quotes.

    @param string: Input string
    @return: Sanitized string
    """
    return sanitize_internal(string, False)


def sanitize_value(string):
    """Sanitize string of tags value and source.

    @oaram string: Input String
    @return: Sanitized String
    """
    res = string.strip()
    res = res.replace("\"", "\\\"")
    return "\"" + res.replace("\n", "\\n") + "\""


def sanitize_internal(string, add_quotes):
    """Sanitize string of metric name, key of tags.

    @param string: Input string
    @add_quotes: Add quotes or not
    @return: Sanitized String
    """
    builder = ''
    if add_quotes:
        builder += '"'
    for i, char in enumerate(string):
        is_legal = True
        cur = ord(char)
        if not (44 <= cur <= 46) and not (48 <= cur <= 57) \
            and not (65 <= cur <= 90) and not \
                (97 <= cur <= 122) and not cur == 95:
            # legal characters are any single character
            # between , (index 44) and . (index 46) or
            # between 0 (index 48) and 9 (index 57) or
            # between A (index 65) and Z (index 90) or
            # between a (index 97) and z (index 122) or
            # _ (index 95)
            if not ((i == 0 and cur == 0x2206) or (i == 0 and cur == 0x0394) or
                    (i == 0 and cur == 126)):
                is_legal = False
            # first character can also be \u2206 (∆ - INCREMENT) or
            # \u0394 (Δ - GREEK CAPITAL LETTER DELTA) or
            # ~ tilda character for internal metrics
        builder += char if is_legal else '-'
    if add_quotes:
        builder += '"'
    return builder


def is_blank(string):
    """
    Check is a string is black or not, either none or only contains whitespace.

    @param string: String to be checked
    @return: Is blank or not
    """
    return string is None or len(string) == 0 or string.isspace()
    # return len(re.sub(r'[\s]+', '', s)) == 0


def validate_tags(tags):
    """
    Ensure that the tag is not empty, otherwise throw the error.

    :param tags:
    :return:
    """
    for tag in tags:
        if is_blank(tag):
            raise ValueError('Event tag cannot be blank')


def validate_annotations(annotations):
    """
    Ensure that annotation key and value are not empty.

    Otherwise throw the error.

    :param annotations:
    :return:
    """
    for key, value in annotations.items():
        if is_blank(key):
            raise ValueError('Annotation key cannot be blank')
        if is_blank(value):
            raise ValueError('Annotation value cannot be blank for key: '
                             + key)


# pylint: disable=too-many-arguments

def metric_to_line_data(name, value, timestamp, source, tags, default_source):
    """Metric Data to String.

    Wavefront Metrics Data format
    <metricName> <metricValue> [<timestamp>] source=<source> [pointTags]
    Example: "new-york.power.usage 42422 1533531013 source=localhost
              datacenter=dc1"

    @param name: Metric Name
    @type name: str
    @param value: Metric Value
    @type value: float
    @param timestamp: Timestamp
    @type timestamp: long
    @param source: Source
    @type source: str
    @param tags: Tags
    @type tags: dict
    @param default_source:
    @type default_source: str
    @return: String
    """
    if is_blank(name):
        raise ValueError('Metrics name cannot be blank')

    if is_blank(source):
        source = default_source

    str_builder = [sanitize(name), str(float(value))]
    if timestamp is not None:
        str_builder.append(str(int(timestamp)))
    str_builder.append('source=' + sanitize_value(source))
    if tags is not None:
        for key, val in tags.items():
            if is_blank(key):
                raise ValueError('Metric point tag key cannot be blank')
            if is_blank(val):
                raise ValueError('Metric point tag value cannot be blank')
            str_builder.append(sanitize(key) + '=' + sanitize_value(val))
    return ' '.join(str_builder) + '\n'


def histogram_to_line_data(name, centroids, histogram_granularities, timestamp,
                           source, tags, default_source):
    """Wavefront Histogram Data format.

    {!M | !H | !D} [<timestamp>] #<count> <mean> [centroids] <histogramName>
    source=<source> [pointTags]
    Example: "!M 1533531013 #20 30.0 #10 5.1 request.latency source=appServer1
              region=us-west"

    @param name: Histogram Name
    @type name: str
    @param centroids: List of centroids(pairs)
    @type centroids: list
    @param histogram_granularities: Histogram Granularities
    @type histogram_granularities: set
    @param timestamp: Timestamp
    @type timestamp: long
    @param source: Source
    @type source: str
    @param tags: Tags
    @type tags: dict
    @param default_source: Default Source
    @type default_source: str
    @return: String data of Histogram
    """
    if is_blank(name):
        raise ValueError('Histogram name cannot be blank')

    if not histogram_granularities:
        raise ValueError('Histogram granularities cannot be null or empty')

    if not centroids:
        raise ValueError('A distribution should have at least one centroid')

    if is_blank(source):
        source = default_source

    line_builder = []
    for histogram_granularity in histogram_granularities:
        str_builder = [histogram_granularity]
        if timestamp is not None:
            str_builder.append(str(int(timestamp)))
        for centroid_1, centroid_2 in centroids:
            str_builder.append('#' + str(centroid_2))
            str_builder.append(str(centroid_1))
        str_builder.append(sanitize(name))
        str_builder.append('source=' + sanitize_value(source))
        if tags is not None:
            for key in tags:
                if is_blank(key):
                    raise ValueError('Histogram tag key cannot be blank')
                if is_blank(tags[key]):
                    raise ValueError('Histogram tag value cannot be blank')
                str_builder.append(
                    sanitize(key) + '=' + sanitize_value(tags[key]))
        line_builder.append(' '.join(str_builder))
    return '\n'.join(line_builder) + '\n'


# pylint: disable=unused-argument
def tracing_span_to_line_data(name, start_millis, duration_millis, source,
                              trace_id, span_id, parents, follows_from, tags,
                              span_logs, default_source):
    """Wavefront Tracing Span Data format.

    <tracingSpanName> source=<source> [pointTags] <start_millis>
    <duration_milli_seconds>
    Example: "getAllUsers source=localhost
              traceId=7b3bf470-9456-11e8-9eb6-529269fb1459
              spanId=0313bafe-9457-11e8-9eb6-529269fb1459
              parent=2f64e538-9457-11e8-9eb6-529269fb1459
              application=Wavefront http.method=GET
              1533531013 343500"

    @param name: Span Name
    @type name: str
    @param start_millis: Start time
    @type start_millis: long
    @param duration_millis: Duration time
    @type duration_millis: long
    @param source: Source
    @type source: str
    @param trace_id: Trace ID
    @type trace_id: UUID
    @param span_id: Span ID
    @type span_id: UUID
    @param parents: Parents Span ID
    @type parents: List of UUID
    @param follows_from: Follows Span ID
    @type follows_from: List of UUID
    @param tags: Tags
    @type tags: list
    @param span_logs: Span Log
    @param default_source: Default Source
    @type default_source: str
    @return: String data of tracing span
    """
    # pylint: disable=too-many-locals
    if is_blank(name):
        raise ValueError('Span name cannot be blank')

    if is_blank(source):
        source = default_source

    str_builder = [sanitize_value(name),
                   'source=' + sanitize_value(source),
                   'traceId=' + str(trace_id),
                   'spanId=' + str(span_id)]
    if parents is not None:
        for uuid in parents:
            str_builder.append('parent=' + str(uuid))
    if follows_from is not None:
        for uuid in follows_from:
            str_builder.append('followsFrom=' + str(uuid))
    if span_logs:
        tags.append((SPAN_LOG_KEY, 'true'))
    if tags is not None:
        tag_set = set()
        for key, val in tags:
            if is_blank(key):
                raise ValueError('Span tag key cannot be blank')
            if is_blank(val):
                raise ValueError('Span tag val cannot be blank')
            cur_tag = sanitize(key) + '=' + sanitize_value(val)
            if cur_tag not in tag_set:
                str_builder.append(cur_tag)
                tag_set.add(cur_tag)
    str_builder.append(str(start_millis))
    str_builder.append(str(duration_millis))
    return ' '.join(str_builder) + '\n'


def span_log_to_line_data(trace_id, span_id, span_logs, scrambler=None):
    """Wavefront Tracing Span Log JSON format.

    @param trace_id: Trace ID
    @param span_id: Span ID
    @param span_logs: Span Log
    @param scrambler: Additional UUID, optional
    @return: Span Log in JSON Format
    """
    span_log_json = {'traceId': str(trace_id),
                     'spanId': str(span_id),
                     'logs': span_logs}
    if scrambler:
        span_log_json['_scrambler'] = str(scrambler)
    return json.dumps(span_log_json, default=lambda o: o.__dict__) + '\n'


# pylint: disable=too-many-branches
def event_to_json(name, start_time, end_time, source, tags,
                  annotations, default_source):
    """Wavefront Event JSON format for direct data ingestion.

    @param name: Event Name
    @type name: str
    @param start_time: Event Start Time
    @type start_time: long
    @param end_time: Event End Time
    @type end_time: long
    @param source: Source
    @type source: str
    @param tags: Tags
    @type tags: list[str]
    @param annotations: Annotations
    @type annotations: dict
    @param default_source: Default Host Name
    @type default_source: str
    @return: Event JSON as String
    """
    if is_blank(name):
        raise ValueError('Event name cannot be blank')

    if not start_time:
        raise ValueError('Event start time cannot be blank')

    if not source:
        source = default_source

    event = {'name': name, 'annotations': {}, 'hosts': [source]}
    event['startTime'] = start_time

    if end_time:
        event['endTime'] = end_time
    else:
        event['endTime'] = start_time + 1

    if annotations:
        validate_annotations(annotations)
        event['annotations'] = annotations

    if tags:
        validate_tags(tags)
        event['tags'] = tags

    return str(json.dumps(event))


def get_sem_ver(name):
    """Return semantic version of sdk used in Wavefront reportable format.

    Ex: <major>.<2-digit-minor><2-digit-patch> (1.0603 => v1.6.3)

    @param name: SDK Name
    @type name: str
    @return: Semantic version in wavefront format as String
    """
    try:
        version = pkg_resources.require(name)[0].version
        return get_sem_ver_value(version)
    except pkg_resources.DistributionNotFound:
        LOGGER.warning('Unable to get version info,'
                       ' No distribution found for : %s', name)
    return "0.0"


def get_sem_ver_value(version):
    """Return semantic version of sdk in Wavefront reportable format.

    Ex: <major>.<2-digit-minor><2-digit-patch> (1.0603 => v1.6.3)

    @param version: SDK Version
    @type version: str
    @return: Semantic version in wavefront format as String
    """
    for match in re.finditer(
            "([0-9]\\d*)\\.(\\d+)\\.(\\d+)(?:-([a-zA-Z0-9]+))?", version):
        major = match.group(1) + "."
        minor = match.group(2) \
            if len(match.group(2)) != 1 else "0" + match.group(2)
        patch = match.group(3) \
            if len(match.group(3)) != 1 else "0" + match.group(3)
        return major + minor + patch
    return "0.0"


def event_to_line_data(name, start_time, end_time, source, tags,
                       annotations, default_source):
    """Wavefront Event Line format for data ingestion via proxy.

    @param name: Event Name
    @type name: str
    @param start_time: Event Start Time
    @type start_time: long
    @param end_time: Event End Time
    @type end_time: long
    @param source: Source
    @type source: str
    @param tags: Tags
    @type tags: list[str]
    @param annotations: Annotations
    @type annotations: dict
    @param default_source: Default Host Name
    @type default_source: str
    @return: Event as String
    """
    str_builder = ['@Event']
    if is_blank(name):
        raise ValueError('Event name cannot be blank')

    if not start_time:
        raise ValueError('Event start time cannot be blank')

    str_builder.append(str(start_time))

    if end_time:
        str_builder.append(str(end_time))
    else:
        str_builder.append(str(start_time + 1))

    str_builder.append('"' + name + '"')

    if annotations:
        validate_annotations(annotations)
        for key, value in annotations.items():
            str_builder.append(key + '="' + value + '"')

    if not source:
        source = default_source

    str_builder.append('host="' + source + '"')

    if tags:
        validate_tags(tags)
        for tag in tags:
            str_builder.append('tag="' + tag + '"')

    return ' '.join(str_builder) + '\n'



# -*- coding: utf-8 -*-
"""Wavefront SDK Gauge.

@author Hao Song (songhao@vmware.com)
"""

from wavefront_sdk.common.metrics import metrics


# pylint: disable=too-few-public-methods
class WavefrontSdkGauge(metrics.WavefrontSdkMetric):
    """Wavefront SDK Gauge."""

    def __init__(self, supplier):
        """Construct Wavefront SDK Gauge."""
        self.supplier = supplier

    def get_value(self):
        """Get value of the gauge."""
        if callable(self.supplier):
            return self.supplier()
        return None



# -*- coding: utf-8 -*-
"""Interface of Metric Sender for both clients.

@author Hao Song (songhao@vmware.com)
"""
from __future__ import unicode_literals


# pylint: disable=E0012,R0205
class WavefrontMetricSender(object):
    """Interface of Metric Sender for both clients."""

    # ∆: INCREMENT
    DELTA_PREFIX = '∆'  # '\u2206'

    # Δ: GREEK CAPITAL LETTER DELTA
    DELTA_PREFIX_2 = 'Δ'  # '\u0394'

    # pylint: disable=too-many-arguments
    def send_metric(self, name, value, timestamp, source, tags):
        """Send Metric Data.

        Wavefront Metrics Data format
        <metricName> <metricValue> [<timestamp>] source=<source> [pointTags]
        Example: "new-york.power.usage 42422 1533531013 source=localhost
                  datacenter=dc1"

        @param name: Metric Name
        @type name: str
        @param value: Metric Value
        @type value: float
        @param timestamp: Timestamp
        @type timestamp: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: dict
        """
        raise NotImplementedError

    def send_formatted_metric(self, point):
        """Send a formatted metric immediately.

        @param point: Formatted metric point
        @type: str
        """
        self.send_metric_now([point])

    def send_metric_now(self, metrics):
        """Send a list of metrics immediately.

        Have to construct the data manually by calling
        common.utils.metric_to_line_data()

        @param metrics: List of string spans data
        @type metrics: list[str]
        """
        raise NotImplementedError

    def send_delta_counter(self, name, value, source, tags):
        """Send Delta Counter Data.

        @param name: Metric Name
        @type name: str
        @param value: Metric Value
        @type value: float
        @param source: Source
        @type source: str
        @param tags: Tags
        @param tags: dict
        """
        if not (name.startswith(self.DELTA_PREFIX) or
                name.startswith(self.DELTA_PREFIX_2)):
            name = self.DELTA_PREFIX + name
        self.send_metric(name, value, None, source, tags)



# -*- coding: utf-8 -*-
"""A connection handler for both proxy and direct client.

@author: Hao Song (songhao@vmware.com)
"""

from . import utils


# pylint: disable=E0012,R0205
class ConnectionHandler(object):
    """A connection handler for both proxy and direct client.

    Handle the failure atomic counter.
    """

    def __init__(self):
        """Initialize the failure atomic counter."""
        self._failure = utils.AtomicCounter()

    def get_failure_count(self):
        """Get failure count for one connection.

        @return: failure count
        """
        return self._failure.value

    def increment_failure_count(self):
        """Increment failure count by one."""
        self._failure.increment()



# -*- coding: utf-8 -*-
"""Interface of Histogram Sender for both clients.

@author Hao Song (songhao@vmware.com)
"""


# pylint: disable=E0012,R0205
class WavefrontHistogramSender(object):
    """Interface of Histogram Sender for both clients."""

    # pylint: disable=too-many-arguments
    def send_distribution(self, name, centroids, histogram_granularities,
                          timestamp, source, tags):
        """Send Distribution Data.

        Wavefront Histogram Data format
        {!M | !H | !D} [<timestamp>] #<count> <mean> [centroids]
        <histogramName> source=<source> [pointTags]
        Example: "!M 1533531013 #20 30.0 #10 5.1 request.latency
                  source=appServer1 region=us-west"

        @param name: Histogram Name
        @type name: str
        @param centroids: List of centroids(pairs)
        @type centroids: list
        @param histogram_granularities: Histogram Granularities
        @type histogram_granularities: set
        @param timestamp: Timestamp
        @type timestamp: long
        @param source: Source
        @type source: str
        @param tags: Tags
        @type tags: dict
        """
        raise NotImplementedError

    def send_formatted_metric(self, distribution):
        """Send a formatted histogram immediately.

        @param distribution: Formatted distribution
        @type: str
        """
        self.send_distribution_now([distribution])

    def send_distribution_now(self, distributions):
        """Send a list of distributions immediately.

        Have to construct the data manually by calling
        common.utils.histogram_to_line_data()

        @param distributions: List of string distribution data
        @type distributions: list[str]
        """
        raise NotImplementedError



"""Unit Tests for Wavefront Application Tags.

@author Shipeng Xie (xshipeng@vmware.com)
"""

import unittest
from unittest.mock import patch

from wavefront_sdk.common.application_tags import ApplicationTags


class TestApplicationTagsImpl(unittest.TestCase):
    """Unit Tests of ApplicationTags."""

    def setUp(self):
        """Initialize for tests."""
        self.env_patch = patch.dict(
            'os.environ', {
                'app_label_1': 'value_1',
                'app_label_2': 'value_2',
                'label_3': 'value_3'
            }, clear=True)
        self.application_tags = ApplicationTags(
            'test_app',
            'test_service',
            custom_tags=[('label_1', 'value_1')])
        self.env_patch.start()

    def tearDown(self):
        """Clean for tests."""
        self.env_patch.stop()

    def test_add_custom_tags_from_env(self):
        """Test add_custom_tags_from_env function."""
        self.assertEqual([('label_1', 'value_1')],
                         self.application_tags.custom_tags)
        self.application_tags.add_custom_tags_from_env(r'^App.*$')
        self.assertCountEqual([('label_1', 'value_1'),
                               ('app_label_1', 'value_1'),
                               ('app_label_2', 'value_2')],
                              self.application_tags.custom_tags)
        self.application_tags.add_custom_tags_from_env('app_tag')
        self.assertCountEqual([('label_1', 'value_1'),
                               ('app_label_1', 'value_1'),
                               ('app_label_2', 'value_2')],
                              self.application_tags.custom_tags)

    def test_add_custom_tag_from_env(self):
        """Test add_custom_tag_from_env function."""
        self.assertEqual([('label_1', 'value_1')],
                         self.application_tags.custom_tags)
        self.application_tags.add_custom_tag_from_env('label_2', 'app_label_2')
        self.assertCountEqual([('label_1', 'value_1'), ('label_2', 'value_2')],
                              self.application_tags.custom_tags)


if __name__ == '__main__':
    # run 'python -m unittest discover' from top-level to run tests
    unittest.main()



#!/usr/bin/env python3
# coding: utf-8
"""Wavefront Python SDK.

<p>This is a Wavefront Python SDK</p>
"""

import os

import setuptools

with open(os.path.join(os.path.abspath(os.path.dirname(__file__)),
                       'README.md')) as fd:
    LONG_DESCRIPTION = fd.read()

setuptools.setup(
    name='wavefront-sdk-python',
    version='1.7.3',
    author='Wavefront by VMware',
    author_email='chitimba@wavefront.com',
    url='https://github.com/wavefrontHQ/wavefront-sdk-python',
    license='Apache-2.0',
    description='Wavefront Python SDK',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    keywords=[
        'Wavefront',
        'Wavefront SDK'
        ],
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
        ],
    include_package_data=True,
    packages=setuptools.find_packages(exclude=('*.tests', '*.tests.*',
                                               'tests.*', 'tests')),
    install_requires=(
        'requests>=2.18.4',
        'tdigest>=0.5.2'
        )
)
